"""Dimmable light states."""
import logging

from insteonplm.constants import (
    COMMAND_LIGHT_BRIGHTEN_ONE_STEP_0X15_0X00,
    COMMAND_LIGHT_DIM_ONE_STEP_0X16_0X00,
    COMMAND_LIGHT_INSTANT_CHANGE_0X21_NONE,
    COMMAND_LIGHT_MANUALLY_TURNED_OFF_0X22_0X00,
    COMMAND_LIGHT_MANUALLY_TURNED_ON_0X23_0X00,
    COMMAND_LIGHT_OFF_0X13_0X00,
    COMMAND_LIGHT_OFF_FAST_0X14_0X00,
    COMMAND_LIGHT_ON_0X11_NONE,
    COMMAND_LIGHT_ON_FAST_0X12_NONE,
    COMMAND_LIGHT_STATUS_REQUEST_0X19_0X00,
    COMMAND_LIGHT_STATUS_REQUEST_0X19_NONE,
    COMMAND_LIGHT_STOP_MANUAL_CHANGE_0X18_0X00,
    FAN_SPEED_HIGH,
    FAN_SPEED_LOW,
    FAN_SPEED_MEDIUM,
    FAN_SPEED_OFF,
    MESSAGE_TYPE_ALL_LINK_BROADCAST,
    MESSAGE_TYPE_ALL_LINK_CLEANUP,
)
from insteonplm.messages.standardSend import StandardSend
from insteonplm.messages.extendedSend import ExtendedSend
from insteonplm.messages.standardReceive import StandardReceive
from insteonplm.messages.messageFlags import MessageFlags
from insteonplm.states import State

_LOGGER = logging.getLogger(__name__)
DIMMABLE_KEYPAD_SCENE_ON_LEVEL = "sceneOnLevel"


class DimmableSwitch(State):
    """Device state representing an On/Off switch that is controllable.

    Available methods are:
    on()
    off()
    set_level()
    brightent()
    dim()
    connect()
    update(self, val)
    async_refresh_state()
    """

    def __init__(
        self,
        address,
        statename,
        group,
        send_message_method,
        message_callbacks,
        defaultvalue=None,
    ):
        """Init the DimmableSwitch Class."""
        super().__init__(
            address,
            statename,
            group,
            send_message_method,
            message_callbacks,
            defaultvalue,
        )

        self._updatemethod = self._send_status_request
        self._register_messages()

    # pylint: disable=too-many-locals
    def _register_messages(self):
        _LOGGER.debug(
            "Registering callbacks for DimmableSwitch device %s", self._address.human
        )
        template_on_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
            cmd2=self._group,
        )
        template_on_fast_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_FAST_0X12_NONE,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
            cmd2=self._group,
        )
        template_off_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_0X13_0X00,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
            cmd2=self._group,
        )
        template_off_fast_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_FAST_0X14_0X00,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
            cmd2=self._group,
        )
        template_manual_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_STOP_MANUAL_CHANGE_0X18_0X00,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
            cmd2=self._group,
        )
        template_instant_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_INSTANT_CHANGE_0X21_NONE,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
            cmd2=self._group,
        )
        template_manual_off_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_MANUALLY_TURNED_OFF_0X22_0X00,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
            cmd2=self._group,
        )
        template_manual_on_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_MANUALLY_TURNED_ON_0X23_0X00,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
            cmd2=self._group,
        )

        template_on_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
        )
        template_on_fast_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_FAST_0X12_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
        )
        template_off_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_0X13_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )
        template_off_fast_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_FAST_0X14_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )
        template_manual_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_STOP_MANUAL_CHANGE_0X18_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )
        template_instant_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_INSTANT_CHANGE_0X21_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )
        template_manual_off_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_MANUALLY_TURNED_OFF_0X22_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )
        template_manual_on_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_MANUALLY_TURNED_ON_0X23_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )

        self._message_callbacks.add(template_on_cleanup, self._manual_change_received)
        self._message_callbacks.add(
            template_on_fast_cleanup, self._manual_change_received
        )
        self._message_callbacks.add(template_off_cleanup, self._manual_change_received)
        self._message_callbacks.add(
            template_off_fast_cleanup, self._manual_change_received
        )
        self._message_callbacks.add(
            template_manual_cleanup, self._manual_change_received
        )
        self._message_callbacks.add(
            template_instant_cleanup, self._manual_change_received
        )
        self._message_callbacks.add(
            template_manual_off_cleanup, self._manual_change_received
        )
        self._message_callbacks.add(
            template_manual_on_cleanup, self._manual_change_received
        )

        self._message_callbacks.add(template_on_broadcast, self._on_message_received)
        self._message_callbacks.add(
            template_on_fast_broadcast, self._on_message_received
        )
        self._message_callbacks.add(template_off_broadcast, self._off_message_received)
        self._message_callbacks.add(
            template_off_fast_broadcast, self._off_message_received
        )
        self._message_callbacks.add(
            template_manual_broadcast, self._manual_change_received
        )
        self._message_callbacks.add(
            template_instant_broadcast, self._manual_change_received
        )
        self._message_callbacks.add(
            template_manual_off_broadcast, self._manual_change_received
        )
        self._message_callbacks.add(
            template_manual_on_broadcast, self._manual_change_received
        )

    def on(self):
        """Turn the device ON."""
        on_command = StandardSend(self._address, COMMAND_LIGHT_ON_0X11_NONE, cmd2=0xFF)
        self._send_method(on_command, self._on_message_received)

    def off(self):
        """Turn the device off."""
        off_command = StandardSend(self._address, COMMAND_LIGHT_OFF_0X13_0X00)
        self._send_method(off_command, self._off_message_received)

    def set_level(self, val):
        """Set the devive ON LEVEL."""
        if val == 0:
            self.off()
        else:
            setlevel = 255
            if val < 1:
                setlevel = val * 100
            elif val <= 0xFF:
                setlevel = val
            set_command = StandardSend(
                self._address, COMMAND_LIGHT_ON_0X11_NONE, cmd2=setlevel
            )
            self._send_method(set_command, self._on_message_received)

    def brighten(self):
        """Brighten the device one step."""
        brighten_command = StandardSend(
            self._address, COMMAND_LIGHT_BRIGHTEN_ONE_STEP_0X15_0X00
        )
        self._send_method(brighten_command)

    def dim(self):
        """Dim the device one step."""
        dim_command = StandardSend(self._address, COMMAND_LIGHT_DIM_ONE_STEP_0X16_0X00)
        self._send_method(dim_command)

    def _on_message_received(self, msg):
        cmd2 = msg.cmd2 if msg.cmd2 else 255
        self._update_subscribers(cmd2)

    # pylint: disable=unused-argument
    def _off_message_received(self, msg):
        self._update_subscribers(0x00)

    # pylint: disable=unused-argument
    def _manual_change_received(self, msg):
        self._send_status_request()

    def _send_status_request(self):
        status_command = StandardSend(
            self._address, COMMAND_LIGHT_STATUS_REQUEST_0X19_0X00
        )
        self._send_method(status_command, self._status_message_received)

    def _status_message_received(self, msg):
        _LOGGER.debug("DimmableSwitch status message received called")
        self._update_subscribers(msg.cmd2)


class DimmableSwitch_Fan(DimmableSwitch):
    """Device state representing a controlable bottom outlet On/Off switch.

    Available methods are:
    on(self)
    off(self)
    async_refresh_state(self)
    connect(self, call_back)
    update(self, val)
    async_refresh_state(self)
    """

    def __init__(
        self,
        address,
        statename,
        group,
        send_message_method,
        set_message_callback_method,
        defaultvalue=None,
    ):
        """Init the DimmableSwitch_Fan Class."""
        super().__init__(
            address,
            statename,
            group,
            send_message_method,
            set_message_callback_method,
            defaultvalue,
        )

        self._updatemethod = self._status_request
        self._udata = {"d1": self._group}

    def on(self):
        """Turn on the fan."""
        on_command = ExtendedSend(
            self._address,
            COMMAND_LIGHT_ON_0X11_NONE,
            self._udata,
            cmd2=FAN_SPEED_MEDIUM,
        )
        on_command.set_checksum()
        self._send_method(on_command, self._on_message_received)

    def set_level(self, val):
        """Set the fan speed."""
        speed = self._value_to_fan_speed(val)
        if val == 0:
            self.off()
        else:
            set_command = ExtendedSend(
                self._address, COMMAND_LIGHT_ON_0X11_NONE, self._udata, cmd2=speed
            )
            set_command.set_checksum()
            self._send_method(set_command, self._on_message_received)

    def off(self):
        """Turn off the fan."""
        off_command = ExtendedSend(
            self._address, COMMAND_LIGHT_OFF_0X13_0X00, self._udata
        )
        off_command.set_checksum()
        self._send_method(off_command, self._off_message_received)
        _LOGGER.debug("Ending DimmableSwitch_Fan.off")

    def _status_request(self):
        status_command = StandardSend(
            self._address, COMMAND_LIGHT_STATUS_REQUEST_0X19_NONE, cmd2=0x03
        )
        self._send_method(status_command, self._status_message_received)

    def _status_message_received(self, msg):
        self._update_subscribers(msg.cmd2)

    # pylint: disable=no-self-use
    def _value_to_fan_speed(self, speed):
        if speed > 0xFE:
            return FAN_SPEED_HIGH
        if speed > 0x7F:
            return FAN_SPEED_MEDIUM
        if speed > 0:
            return FAN_SPEED_LOW
        return FAN_SPEED_OFF


class DimmableRemote(State):
    """Device state representing dimmable keypad switch that is not controllable.

    Available methods are:
    connect()
    update(self, val)
    async_refresh_state()
    """

    def __init__(
        self,
        address,
        statename,
        group,
        send_message_method,
        message_callbacks,
        defaultvalue=None,
    ):
        """Init the DimmableSwitch Class."""
        super().__init__(
            address,
            statename,
            group,
            send_message_method,
            message_callbacks,
            defaultvalue,
        )

        _LOGGER.debug(
            "Registering callbacks for DimmableSwitch device %s", self._address.human
        )

        self._is_responder = False

        self._register_messages()

    def _on_message_received(self, msg):
        _LOGGER.debug(
            "Calling DimmableRemote _on_message_received " "for group %d with level %d",
            self._group,
            msg.cmd2,
        )
        self._update_subscribers(0xFF)

    # pylint: disable=unused-argument
    def _off_message_received(self, msg):
        _LOGGER.debug(
            "Calling DimmableRemote _off_message_received " "for group %d", self._group
        )
        self._update_subscribers(0x00)

    # pylint: disable=no-self-use
    # pylint: disable=unused-argument
    def _manual_change_received(self, msg):
        _LOGGER.debug(
            "Message type 0x17 or 0x18 received but no way "
            "to properly handle them, sorry."
        )

    def _register_messages(self):
        template_on_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
        )
        template_on_fast_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_FAST_0X12_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
        )
        template_off_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_0X13_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )
        template_off_fast_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_FAST_0X14_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )
        template_manual_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_STOP_MANUAL_CHANGE_0X18_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )
        template_instant_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_INSTANT_CHANGE_0X21_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )
        template_manual_off_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_MANUALLY_TURNED_OFF_0X22_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )
        template_manual_on_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_MANUALLY_TURNED_ON_0X23_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )

        self._message_callbacks.add(template_on_broadcast, self._on_message_received)
        self._message_callbacks.add(
            template_on_fast_broadcast, self._on_message_received
        )
        self._message_callbacks.add(template_off_broadcast, self._off_message_received)
        self._message_callbacks.add(
            template_off_fast_broadcast, self._off_message_received
        )
        self._message_callbacks.add(
            template_manual_broadcast, self._manual_change_received
        )
        self._message_callbacks.add(
            template_instant_broadcast, self._manual_change_received
        )
        self._message_callbacks.add(
            template_manual_off_broadcast, self._manual_change_received
        )
        self._message_callbacks.add(
            template_manual_on_broadcast, self._manual_change_received
        )


class DimmableKeypadA(DimmableSwitch):
    """Device state for a controllable keypad button A On/Off switch."""

    def __init__(
        self,
        address,
        statename,
        group,
        send_message_method,
        message_callbacks,
        defaultvalue,
        leds,
    ):
        """Init the DimmableKeypadA class."""
        super().__init__(
            address,
            statename,
            group,
            send_message_method,
            message_callbacks,
            defaultvalue,
        )

        self._updatemethod = self._send_status_request
        self._leds = leds

    def led_on(self):
        """Turn the LED on."""
        self._leds.on(self._group)

    def led_off(self):
        """Turn the LED off."""
        self._leds.off(self._group)

    def led_is_on(self):
        """Return if the LED is on."""
        return self._leds.is_on(self._group)

    def _on_message_received(self, msg):
        super()._on_message_received(msg)
        self._leds.async_refresh_state()

    def _off_message_received(self, msg):
        super()._off_message_received(msg)
        self._leds.async_refresh_state()

    def _send_status_request(self):
        switch_status_msg = StandardSend(
            self._address, COMMAND_LIGHT_STATUS_REQUEST_0X19_0X00
        )
        self._send_method(switch_status_msg, self._status_message_received)



"""On/Off states."""
import asyncio
import logging

import async_timeout

from insteonplm.constants import (
    COMMAND_LIGHT_MANUALLY_TURNED_ON_0X23_0X00,
    COMMAND_LIGHT_MANUALLY_TURNED_OFF_0X22_0X00,
    COMMAND_LIGHT_OFF_0X13_0X00,
    COMMAND_LIGHT_OFF_FAST_0X14_0X00,
    COMMAND_LIGHT_ON_0X11_NONE,
    COMMAND_LIGHT_ON_FAST_0X12_NONE,
    COMMAND_LIGHT_STATUS_REQUEST_0X19_0X00,
    COMMAND_LIGHT_STATUS_REQUEST_0X19_0X01,
    MESSAGE_TYPE_ALL_LINK_BROADCAST,
    MESSAGE_TYPE_ALL_LINK_CLEANUP,
    COMMAND_EXTENDED_TRIGGER_ALL_LINK_0X30_0X00,
    COMMAND_EXTENDED_GET_SET_0X2E_0X00,
)
from insteonplm.messages.standardSend import StandardSend
from insteonplm.messages.extendedSend import ExtendedSend
from insteonplm.messages.standardReceive import StandardReceive
from insteonplm.messages.extendedReceive import ExtendedReceive
from insteonplm.messages.messageFlags import MessageFlags
from insteonplm.messages.userdata import Userdata
from insteonplm.states import State
from insteonplm.utils import bit_is_set, set_bit

_LOGGER = logging.getLogger(__name__)
DIMMABLE_KEYPAD_SCENE_ON_LEVEL = "scene_on_level"


class OnOffStateBase(State):
    """Base state representing an On/Off switch that is controllable.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
      register_updates()
      async_refresh_state()
    """

    def __init__(
        self,
        address,
        statename,
        group,
        send_message_method,
        message_callbacks,
        defaultvalue=None,
    ):
        """Init the OnOffStateBase."""
        super().__init__(
            address,
            statename,
            group,
            send_message_method,
            message_callbacks,
            defaultvalue,
        )

        self._updatemethod = self._send_status_request

        self._register_messages()

    # pylint: disable=unused-argument
    def _on_message_received(self, msg):
        """Receive a ON message."""
        self._update_subscribers(0xFF)

    # pylint: disable=unused-argument
    def _off_message_received(self, msg):
        """Receive An OFF message."""
        self._update_subscribers(0x00)

    # pylint: disable=unused-argument
    def _manual_change_received(self, msg):
        """Receive a manual change message."""
        self._send_status_request()

    def _send_status_request(self):
        """Send a status request message to the device."""
        status_command = StandardSend(
            self._address, COMMAND_LIGHT_STATUS_REQUEST_0X19_0X00
        )
        self._send_method(status_command, self._status_message_received)

    def _status_message_received(self, msg):
        """Receive a status message."""
        if msg.cmd2 == 0x00:
            self._update_subscribers(0x00)
        else:
            self._update_subscribers(0xFF)

    def _register_messages(self):
        """Register messages to listen for."""
        template_on_group = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
        )
        template_fast_on_group = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_FAST_0X12_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
        )
        template_off_group = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_0X13_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )
        template_fast_off_group = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_FAST_0X14_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )
        template_manual_on_group = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_MANUALLY_TURNED_ON_0X23_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )
        template_manual_off_group = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_MANUALLY_TURNED_OFF_0X22_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )

        template_on_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
            cmd2=self._group,
        )
        template_fast_on_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_FAST_0X12_NONE,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
            cmd2=self._group,
        )
        template_off_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_0X13_0X00,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
            cmd2=self._group,
        )
        template_fast_off_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_FAST_0X14_0X00,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
            cmd2=self._group,
        )
        template_manual_on_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_MANUALLY_TURNED_ON_0X23_0X00,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
            cmd2=self._group,
        )
        template_manual_off_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_MANUALLY_TURNED_OFF_0X22_0X00,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
            cmd2=self._group,
        )

        self._message_callbacks.add(template_on_group, self._on_message_received)
        self._message_callbacks.add(template_fast_on_group, self._on_message_received)
        self._message_callbacks.add(template_off_group, self._off_message_received)
        self._message_callbacks.add(template_fast_off_group, self._off_message_received)
        self._message_callbacks.add(
            template_manual_on_group, self._manual_change_received
        )
        self._message_callbacks.add(
            template_manual_off_group, self._manual_change_received
        )

        self._message_callbacks.add(template_on_cleanup, self._on_message_received)
        self._message_callbacks.add(template_fast_on_cleanup, self._on_message_received)
        self._message_callbacks.add(template_off_cleanup, self._off_message_received)
        self._message_callbacks.add(
            template_fast_off_cleanup, self._off_message_received
        )
        self._message_callbacks.add(
            template_manual_on_cleanup, self._manual_change_received
        )
        self._message_callbacks.add(
            template_manual_off_cleanup, self._manual_change_received
        )


class OnOffSwitch(OnOffStateBase):
    """On/Off state representing an On/Off switch that is controllable.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
      on()
      off()
      register_updates()
      async_refresh_state()
    """

    def on(self):
        """Send ON command to device."""
        on_command = StandardSend(self._address, COMMAND_LIGHT_ON_0X11_NONE, 0xFF)
        self._send_method(on_command, self._on_message_received)

    def off(self):
        """Send OFF command to device."""
        off_command = StandardSend(self._address, COMMAND_LIGHT_OFF_0X13_0X00)
        self._send_method(off_command, self._off_message_received)


class OnOffSwitch_OutletTop(OnOffStateBase):
    """Device state representing a controllable top outlet On/Off switch.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
      on()
      off()
      register_updates()
      async_refresh_state()
    """

    def __init__(
        self,
        address,
        statename,
        group,
        send_message_method,
        message_callbacks,
        defaultvalue=None,
    ):
        """Init the OnOffSwitch_OutletTop Class."""
        super().__init__(
            address,
            statename,
            group,
            send_message_method,
            message_callbacks,
            defaultvalue,
        )

        self._updatemethod = self._send_status_0x01_request

    def on(self):
        """Send ON command to device."""
        on_command = StandardSend(self._address, COMMAND_LIGHT_ON_0X11_NONE, 0xFF)
        self._send_method(on_command, self._on_message_received)

    def off(self):
        """Send OFF command to device."""
        self._send_method(
            StandardSend(self._address, COMMAND_LIGHT_OFF_0X13_0X00),
            self._off_message_received,
        )

    def _send_status_0x01_request(self):
        """Sent status request to device."""
        status_command = StandardSend(
            self._address, COMMAND_LIGHT_STATUS_REQUEST_0X19_0X01
        )
        self._send_method(status_command, self._status_message_0x01_received)

    def _status_message_0x01_received(self, msg):
        """Handle status received messages.

        The following status values can be received:
            0x00 = Both Outlets Off
            0x01 = Only Top Outlet On
            0x02 = Only Bottom Outlet On
            0x03 = Both Outlets On
        """
        if msg.cmd2 == 0x00 or msg.cmd2 == 0x02:
            self._update_subscribers(0x00)
        elif msg.cmd2 == 0x01 or msg.cmd2 == 0x03:
            self._update_subscribers(0xFF)
        else:
            raise ValueError


class OnOffSwitch_OutletBottom(OnOffStateBase):
    """Device state representing a controllable bottom outlet On/Off switch.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
      on()
      off()
      register_updates()
      async_refresh_state()
    """

    def __init__(
        self,
        address,
        statename,
        group,
        send_message_method,
        set_message_callback_method,
        defaultvalue=None,
    ):
        """Init the OnOffSwitch_OutletBottom."""
        super().__init__(
            address,
            statename,
            group,
            send_message_method,
            set_message_callback_method,
            defaultvalue,
        )

        self._updatemethod = self._send_status_0x01_request
        self._udata = {"d1": self._group}

    def on(self):
        """Send an ON message to device group."""
        on_command = ExtendedSend(
            self._address, COMMAND_LIGHT_ON_0X11_NONE, self._udata, cmd2=0xFF
        )
        on_command.set_checksum()
        self._send_method(on_command, self._on_message_received)

    def off(self):
        """Send an OFF message to device group."""
        off_command = ExtendedSend(
            self._address, COMMAND_LIGHT_OFF_0X13_0X00, self._udata
        )
        off_command.set_checksum()
        self._send_method(off_command, self._off_message_received)

    def _send_status_0x01_request(self):
        """Send a status request."""
        status_command = StandardSend(
            self._address, COMMAND_LIGHT_STATUS_REQUEST_0X19_0X01
        )
        self._send_method(status_command, self._status_message_received)

    def _status_message_received(self, msg):
        """Receive a status message.

        The following status values can be received:
            0x00 = Both Outlets Off
            0x01 = Only Top Outlet On
            0x02 = Only Bottom Outlet On
            0x03 = Both Outlets On
        """
        if msg.cmd2 == 0x00 or msg.cmd2 == 0x01:
            self._update_subscribers(0x00)
        elif msg.cmd2 == 0x02 or msg.cmd2 == 0x03:
            self._update_subscribers(0xFF)
        else:
            raise ValueError


class OpenClosedRelay(State):
    """Device state representing an Open/Close switch that is controllable.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
      open()
      close()
      register_updates()
      async_refresh_state()
    """

    def __init__(
        self,
        address,
        statename,
        group,
        send_message_method,
        message_callbacks,
        defaultvalue=None,
    ):
        """Init the OpenClosedRelay."""
        super().__init__(
            address,
            statename,
            group,
            send_message_method,
            message_callbacks,
            defaultvalue,
        )

        self._updatemethod = self._send_status_request

    def open(self):
        """Send OPEN command to device."""
        open_command = StandardSend(self._address, COMMAND_LIGHT_ON_0X11_NONE, 0xFF)
        self._send_method(open_command, self._open_message_received)

    def close(self):
        """Send CLOSE command to device."""
        close_command = StandardSend(self._address, COMMAND_LIGHT_OFF_0X13_0X00)
        self._send_method(close_command, self._close_message_received)

    # pylint: disable=unused-argument
    def _open_message_received(self, msg):
        """Receive an OPEN message."""
        self._update_subscribers(0xFF)

    # pylint: disable=unused-argument
    def _close_message_received(self, msg):
        """Receive a CLOSE message."""
        self._update_subscribers(0x00)

    def _send_status_request(self):
        """Send a status request message to the device."""
        status_command = StandardSend(
            self._address, COMMAND_LIGHT_STATUS_REQUEST_0X19_0X00
        )
        self._send_method(status_command, self._status_message_received)

    def _status_message_received(self, msg):
        """Receive a status message."""
        if msg.cmd2 == 0x00:
            self._update_subscribers(0x00)
        else:
            self._update_subscribers(0xFF)


class OnOffKeypadA(OnOffSwitch):
    """Device state for a controllable keypad button A On/Off switch."""

    def __init__(
        self,
        address,
        statename,
        group,
        send_message_method,
        message_callbacks,
        defaultvalue,
        leds,
    ):
        """Init the OnOffKeypadA class."""
        super().__init__(
            address,
            statename,
            group,
            send_message_method,
            message_callbacks,
            defaultvalue,
        )

        self._updatemethod = self._send_status_request
        self._leds = leds

    def led_on(self):
        """Turn the LED is on."""
        self._leds.on(self._group)

    def led_off(self):
        """Turn the LED off."""
        self._leds.off(self._group)

    def led_is_on(self):
        """Return if the LED is on."""
        return self._leds.is_on(self._group)

    def _on_message_received(self, msg):
        super()._on_message_received(msg)
        self._leds.async_refresh_state()

    def _off_message_received(self, msg):
        super()._off_message_received(msg)
        self._leds.async_refresh_state()

    def _send_status_request(self):
        switch_status_msg = StandardSend(
            self._address, COMMAND_LIGHT_STATUS_REQUEST_0X19_0X00
        )
        self._send_method(switch_status_msg, self._status_message_received)


# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-public-methods
class OnOffKeypad(OnOffStateBase):
    """Device state representing a controllable keypad button On/Off switch."""

    def __init__(
        self,
        address,
        statename,
        group,
        send_message_method,
        set_message_callback_method,
        defaultvalue=None,
        loop=None,
        leds=None,
    ):
        """Init the OnOffKeypad Class."""
        super().__init__(
            address,
            statename,
            group,
            send_message_method,
            set_message_callback_method,
            defaultvalue,
        )

        self._leds = leds

        self._on_mask = 0
        self._off_mask = 0
        self._x10_house_code = 20  # 20 means none set
        self._x10_unit = 20  # 20 means none set
        self._ramp_rate = 0
        self._on_level = 0
        self._led_brightness = 0
        self._non_toggle_mask = 0
        self._led_bit_mask = 0
        self._x10_all_bit_mask = 0
        self._on_off_bit_mask = 0
        self._trigger_group_bit_mask = 0

        self._updatemethod = self._leds.async_refresh_state
        self._sent_property = {}
        self._loop = loop
        self._status_response_lock = asyncio.Lock(loop=self._loop)
        self._status_retries = 0
        self._status_received = False

    @property
    def on_mask(self):
        """Return button on mask."""
        return self._on_mask

    @property
    def off_mask(self):
        """Return button on mask."""
        return self._off_mask

    @property
    def x10_house_code(self):
        """Return button on mask."""
        return self._x10_house_code

    @property
    def x10_unit(self):
        """Return button on mask."""
        return self._x10_unit

    @property
    def ramp_rate(self):
        """Return button on mask."""
        return self._ramp_rate

    @property
    def led_brightness(self):
        """Return button on mask."""
        return self._led_brightness

    @property
    def non_toggle_mask(self):
        """Return button on mask."""
        return self._non_toggle_mask

    @property
    def x10_all_bit_mask(self):
        """Return button on mask."""
        return self._x10_all_bit_mask

    @property
    def on_off_bit_mask(self):
        """Return button on mask."""
        return self._on_off_bit_mask

    def on(self):
        """Turn on the LED for the button."""
        self._leds.on(self._group)

    def off(self):
        """Turn off the LED for the button."""
        self._leds.off(self._group)

    def led_is_on(self):
        """Return if the LED is on for the button."""
        return self._leds.is_on(self._group)

    # pylint: disable=unused-argument
    def led_changed(self, addr, group, val):
        """Capture a change to the LED for this button."""
        _LOGGER.debug(
            "Button %d LED changed from %d to %d", self._group, self._value, val
        )
        led_on = bool(val)
        if led_on != bool(self._value):
            self._update_subscribers(int(led_on))

    def set_on_mask(self, mask):
        """Set the on mask for the current group/button."""
        set_cmd = self._create_set_property_msg("_on_mask", 0x02, mask)
        self._send_method(set_cmd, self._property_set)

    def set_off_mask(self, mask):
        """Set the off mask for the current group/button."""
        set_cmd = self._create_set_property_msg("_off_mask", 0x03, mask)
        self._send_method(set_cmd, self._property_set)

    def set_x10_address(self, x10address):
        """Set the X10 address for the current group/button."""
        set_cmd = self._create_set_property_msg("_x10_house_code", 0x04, x10address)
        self._send_method(set_cmd, self._property_set)

    def set_ramp_rate(self, ramp_rate):
        """Set the X10 address for the current group/button."""
        set_cmd = self._create_set_property_msg("_ramp_rate", 0x05, ramp_rate)
        self._send_method(set_cmd, self._property_set)

    def set_on_level(self, val):
        """Set on level for the button/group."""
        on_cmd = self._create_set_property_msg("_on_level", 0x06, val)
        self._send_method(on_cmd, self._property_set)
        self._send_method(on_cmd, self._on_message_received)

    def set_led_brightness(self, brightness):
        """Set the LED brightness for the current group/button."""
        set_cmd = self._create_set_property_msg("_led_brightness", 0x07, brightness)
        self._send_method(set_cmd, self._property_set)

    def set_non_toggle_mask(self, non_toggle_mask):
        """Set the non_toggle_mask for the current group/button."""
        set_cmd = self._create_set_property_msg(
            "_non_toggle_mask", 0x08, non_toggle_mask
        )
        self._send_method(set_cmd, self._property_set)

    def set_x10_all_bit_mask(self, x10_all_bit_mask):
        """Set the x10_all_bit_mask for the current group/button."""
        set_cmd = self._create_set_property_msg(
            "_x10_all_bit_mask", 0x0A, x10_all_bit_mask
        )
        self._send_method(set_cmd, self._property_set)

    def set_trigger_group_bit_mask(self, trigger_group_bit_mask):
        """Set the trigger_group_bit_mask for the current group/button."""
        set_cmd = self._create_set_property_msg(
            "_trigger_group_bit_mask", 0x0C, trigger_group_bit_mask
        )
        self._send_method(set_cmd, self._property_set)

    def scene_on(self):
        """Trigger group/scene to ON level."""
        user_data = Userdata(
            {
                "d1": self._group,
                "d2": 0x00,
                "d3": 0x00,
                "d4": 0x11,
                "d5": 0xFF,
                "d6": 0x00,
            }
        )
        self._set_sent_property(DIMMABLE_KEYPAD_SCENE_ON_LEVEL, 0xFF)
        cmd = ExtendedSend(
            self._address, COMMAND_EXTENDED_TRIGGER_ALL_LINK_0X30_0X00, user_data
        )
        cmd.set_checksum()
        _LOGGER.debug(
            "Calling scene_on and sending response to " "_received_scene_triggered"
        )
        self._send_method(cmd, self._received_scene_triggered)

    def scene_off(self):
        """Trigger group/scene to OFF level."""
        user_data = Userdata(
            {
                "d1": self._group,
                "d2": 0x00,
                "d3": 0x00,
                "d4": 0x13,
                "d5": 0x00,
                "d6": 0x00,
            }
        )
        self._set_sent_property(DIMMABLE_KEYPAD_SCENE_ON_LEVEL, 0x00)
        cmd = ExtendedSend(
            self._address, COMMAND_EXTENDED_TRIGGER_ALL_LINK_0X30_0X00, user_data
        )
        cmd.set_checksum()
        self._send_method(cmd, self._received_scene_triggered)

    def scene_level(self, level):
        """Trigger group/scene to input level."""
        if level == 0:
            self.scene_off()
        else:
            user_data = Userdata(
                {
                    "d1": self._group,
                    "d2": 0x00,
                    "d3": 0x00,
                    "d4": 0x11,
                    "d5": level,
                    "d6": 0x00,
                }
            )
            self._set_sent_property(DIMMABLE_KEYPAD_SCENE_ON_LEVEL, level)
            cmd = ExtendedSend(
                self._address, COMMAND_EXTENDED_TRIGGER_ALL_LINK_0X30_0X00, user_data
            )
            cmd.set_checksum()
            self._send_method(cmd, self._received_scene_triggered)

    def extended_status_request(self):
        """Send status request for group/button."""
        self._status_received = False
        user_data = Userdata({"d1": self.group, "d2": 0x00})
        cmd = ExtendedSend(
            self._address, COMMAND_EXTENDED_GET_SET_0X2E_0X00, userdata=user_data
        )
        cmd.set_checksum()
        self._send_method(cmd, self._status_message_received, True)

    # pylint: disable=unused-argument
    def _on_message_received(self, msg):
        if not self.led_is_on():
            _LOGGER.debug("LED is off and button was turned on")
            self._update_subscribers(1)
            self._leds.manual_on(self._group)
        else:
            _LOGGER.debug("LED is already on when button turned on?")

    # pylint: disable=unused-argument
    def _off_message_received(self, msg):
        if self.led_is_on():
            _LOGGER.debug("LED is on and button was turned off")
            self._update_subscribers(0)
            self._leds.manual_off(self._group)
        else:
            _LOGGER.debug("LED is already off when button turned off?")

    # pylint: disable=unused-argument
    def _manual_change_received(self, msg):
        self._updatemethod()

    # pylint: disable=unused-argument
    def _status_message_received(self, msg):
        """Receive confirmation that the status message is coming.

        The real status message is the extended direct message.
        """
        if not self._status_received:
            asyncio.ensure_future(self._confirm_status_received(), loop=self._loop)

    async def _confirm_status_received(self):
        _LOGGER.debug("Confirming actual status is received")
        if self._status_received:
            _LOGGER.debug("Status was received")
            return

        await self._status_response_lock
        try:
            with async_timeout.timeout(2):
                await self._status_response_lock
                _LOGGER.debug("Actual status received")
        except asyncio.TimeoutError:
            _LOGGER.debug("No status message received")
            if self._status_retries < 10:
                _LOGGER.debug("Resending request")
                self._status_retries += 1
                self.extended_status_request()
            else:
                _LOGGER.debug("Too many retries")
                self._status_retries = 0
        if self._status_response_lock.locked():
            self._status_response_lock.release()

    def _status_extended_message_received(self, msg):
        """Receeive an extended status message.

        Status message received:
            cmd1:  0x2e
            cmd2:  0x00
            flags: Direct Extended
            d1:    group
            d2:    0x01
            d3:    On Mask
            d4:    Off Mask
            d5:    X10 House Code
            d6:    X10 Unit
            d7:    Ramp Rate
            d8:    On-Level
            d9:    LED Brightness
            d10:   Non-Toggle Mask
            d11:   LED Bit Mask
            d12:   X10 ALL Bit Mask
            d13:   On/Off Bit Mask
            d14:   Check sum
        """
        self._status_received = True
        self._status_retries = 0
        _LOGGER.debug("Extended status message received")
        if self._status_response_lock.locked():
            self._status_response_lock.release()
        user_data = msg.userdata
        # self._update_subscribers(user_data['d8'])
        self._set_status_data(user_data)

    def _property_set(self, msg):
        """Set command received and acknowledged."""
        prop = self._sent_property.get("prop")
        if prop and hasattr(self, prop):
            setattr(self, prop, self._sent_property.get("val"))
        self._sent_property = {}

    def _received_scene_triggered(self, msg):
        scene_level = self._sent_property.get("prop")
        # val = self._sent_property.get('val')
        _LOGGER.debug(
            "Calling DimmableKeypad _received_scene_triggered "
            "for group %s with on level %s",
            self._group,
            scene_level,
        )
        if scene_level == DIMMABLE_KEYPAD_SCENE_ON_LEVEL:
            # self._update_subscribers(val)
            self._leds.manual_on(self._group)

    def _set_status_data(self, userdata):
        """Set status properties from userdata response.

        Response values:
            d3:  On Mask
            d4:  Off Mask
            d5:  X10 House Code
            d6:  X10 Unit
            d7:  Ramp Rate
            d8:  On-Level
            d9:  LED Brightness
            d10: Non-Toggle Mask
            d11: LED Bit Mask
            d12: X10 ALL Bit Mask
            d13: On/Off Bit Mask
        """
        self._on_mask = userdata["d3"]
        self._off_mask = userdata["d4"]
        self._x10_house_code = userdata["d5"]
        self._x10_unit = userdata["d6"]
        self._ramp_rate = userdata["d7"]
        self._on_level = userdata["d8"]
        self._led_brightness = userdata["d9"]
        self._non_toggle_mask = userdata["d10"]
        self._led_bit_mask = userdata["d11"]
        self._x10_all_bit_mask = userdata["d12"]
        self._on_off_bit_mask = userdata["d13"]
        self._trigger_group_bit_mask = userdata["d14"]

    def _register_messages(self):
        super()._register_messages()
        template_status_recd = ExtendedReceive.template(
            address=self._address,
            commandtuple=COMMAND_EXTENDED_GET_SET_0X2E_0X00,
            # flags=MessageFlags.template(MESSAGE_TYPE_DIRECT_MESSAGE,
            #                             True),
            userdata={"d1": self._group, "d2": 0x01},
        )
        self._message_callbacks.add(
            template_status_recd, self._status_extended_message_received
        )

    def _create_set_property_msg(self, prop, cmd, val):
        """Create an extended message to set a property.

        Create an extended message with:
            cmd1: 0x2e
            cmd2: 0x00
            flags: Direct Extended
            d1: group
            d2: cmd
            d3: val
            d4 - d14: 0x00

        Parameters:
            prop: Property name to update
            cmd: Command value
                0x02: on mask
                0x03: off mask
                0x04: x10 house code
                0x05: ramp rate
                0x06: on level
                0x07: LED brightness
                0x08: Non-Toggle mask
                0x09: LED bit mask (Do not use in this class. Use LED class)
                0x0a: X10 All bit mask
                0x0c: Trigger group bit mask
            val: New property value

        """
        user_data = Userdata({"d1": self.group, "d2": cmd, "d3": val})
        msg = ExtendedSend(self._address, COMMAND_EXTENDED_GET_SET_0X2E_0X00, user_data)
        msg.set_checksum()
        self._set_sent_property(prop, val)
        return msg

    def _set_sent_property(self, prop, val):
        self._sent_property = {"prop": prop, "val": val}


class OnOffKeypadLed(State):
    """Device state for KeyPadLinc LED."""

    def __init__(
        self,
        address,
        statename,
        group,
        send_message_method,
        message_callbacks,
        defaultvalue=None,
        loop=None,
    ):
        """Init the OnOffKeypadLed class."""
        super().__init__(
            address,
            statename,
            group,
            send_message_method,
            message_callbacks,
            defaultvalue,
        )

        self._updatemethod = self._send_status_request
        self._button_status_method = None

        self._loop = loop
        self._send_led_change_lock = asyncio.Lock(loop=self._loop)
        self._button_observer_callbacks = {}
        self._new_value = 0

    def on(self, group):
        """Turn the LED on for a group."""
        asyncio.ensure_future(self._send_led_on_off_request(group, 1), loop=self._loop)

    def off(self, group):
        """Turn the LED off for a group."""
        asyncio.ensure_future(self._send_led_on_off_request(group, 0), loop=self._loop)

    def manual_on(self, group):
        """Turn the LED on."""
        self._set_led_value(group, 1)
        self._send_status_request()

    def manual_off(self, group):
        """Turn the LED off."""
        self._set_led_value(group, 0)
        self._send_status_request()

    def is_on(self, group):
        """Return if the LED for a button/group is on."""
        val = self._value & 1 << group - 1
        return bool(val)

    def register_led_updates(self, callback, button):
        """Register a callback when a specific button LED changes."""
        button_callbacks = self._button_observer_callbacks.get(button)
        if not button_callbacks:
            self._button_observer_callbacks[button] = []
        _LOGGER.debug("New callback for button %d", button)
        self._button_observer_callbacks[button].append(callback)

    async def _send_led_on_off_request(self, group, val):
        _LOGGER.debug("OnOffKeypadLed._send_led_on_off_request was called")
        await self._send_led_change_lock
        self._new_value = set_bit(self._value, group, bool(val))

        user_data = Userdata({"d1": 0x01, "d2": 0x09, "d3": self._new_value})
        msg = ExtendedSend(self._address, COMMAND_EXTENDED_GET_SET_0X2E_0X00, user_data)
        msg.set_checksum()
        self._send_method(msg, self._on_off_ack_received, True)

    # pylint: disable=unused-argument
    def _on_off_ack_received(self, msg):
        self._update_subscribers(self._new_value)
        if self._send_led_change_lock.locked():
            self._send_led_change_lock.release()

    def _send_status_request(self):
        led_status_msg = StandardSend(
            self._address, COMMAND_LIGHT_STATUS_REQUEST_0X19_0X01
        )
        self._send_method(led_status_msg, self._status_message_received)

    def _status_message_received(self, msg):
        _LOGGER.debug(
            "OnOffKeypadLed status message received with value " "0x%02x", msg.cmd2
        )
        _LOGGER.debug("Status message: %s", msg)
        self._update_subscribers(msg.cmd2)

    def _set_led_value(self, group, val):
        """Set the LED value and confirm with a status check."""
        new_bitmask = set_bit(self._value, group, bool(val))
        self._set_led_bitmask(new_bitmask)

    def _set_led_bitmask(self, bitmask):
        self._value = bitmask

    def _bit_value(self, group, val):
        """Set the LED on/off value from the LED bitmap."""
        bitshift = group - 1
        if val:
            new_value = self._value | (1 << bitshift)
        else:
            new_value = self._value & (0xFF & ~(1 << bitshift))
        return new_value

    def _update_subscribers(self, val):
        old_value = self._value
        self._value = val
        for button in range(1, 9):
            old_bit_set = bit_is_set(old_value, button)
            new_bit_set = bit_is_set(self._value, button)
            old = "on" if old_bit_set else "off"
            new = "on" if new_bit_set else "off"
            _LOGGER.debug("Button %d was %s now is %s", button, old, new)
            # if old_bit_set != new_bit_set:
            callbacks = self._button_observer_callbacks.get(button)
            if callbacks:
                for callback in callbacks:
                    _LOGGER.debug("Calling button update callback %s", callback)
                    callback(self._address, button, int(new_bit_set))
            else:
                _LOGGER.debug("No callbacks found for button %d", button)



"""Module to maintain PLM state information and network interface."""

import asyncio
import logging
import binascii
from collections import deque, namedtuple

import async_timeout

import insteonplm.messages
from insteonplm.constants import (
    MESSAGE_ACK,
    MESSAGE_NAK,
    X10CommandType,
    X10_COMMAND_ALL_UNITS_OFF,
    X10_COMMAND_ALL_LIGHTS_ON,
    X10_COMMAND_ALL_LIGHTS_OFF,
)
from insteonplm.address import Address
from insteonplm.devices import Device, ALDBRecord, ALDBStatus
from insteonplm.linkedDevices import LinkedDevices
from insteonplm.messagecallback import MessageCallback
from insteonplm.messages.allLinkRecordResponse import AllLinkRecordResponse
from insteonplm.messages.getFirstAllLinkRecord import GetFirstAllLinkRecord
from insteonplm.messages.getIMInfo import GetImInfo
from insteonplm.messages.getNextAllLinkRecord import GetNextAllLinkRecord
from insteonplm.messages.manageAllLinkRecord import ManageAllLinkRecord
from insteonplm.messages.setImConfiguration import SetIMConfiguration
from insteonplm.messages.startAllLinking import StartAllLinking
from insteonplm.messages.x10received import X10Received
from insteonplm.messages.x10send import X10Send
from insteonplm.utils import (
    byte_to_housecode,
    byte_to_unitcode,
    rawX10_to_bytes,
    x10_command_type,
)

__all__ = "PLM, Hub"
_LOGGER = logging.getLogger(__name__)
WAIT_TIMEOUT = 1.5
ACKNAK_TIMEOUT = 2


MessageInfo = namedtuple("MessageInfo", "msg wait_nak wait_timeout")


# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-public-methods
class IM(Device, asyncio.Protocol):
    """Handle the Insteon PLM IP control protocol.

    This class is expected to be wrapped inside a Connection class object
    which will maintain the socket and handle auto-reconnects.

    Parameters:
        connection_lost_callback: (optional, callable) called when connection
        is lost to device as defined by asyncio.Protocol

        loop: (optional, asyncio.loop) asyncio event loop

        workdir: (optional, string) Working directory name to save device
        information

        poll_devices: (optional, bool) indicates if the modem should poll the
        devices for status after startup, default is True

        load_aldb; (optional, bool) indicates if the modem should load the
        All-Link Database on startup, default is True

    """

    def __init__(
        self,
        loop=None,
        connection_lost_callback=None,
        workdir=None,
        poll_devices=True,
        load_aldb=True,
    ):
        """Protocol handler that handles all status and changes on PLM."""
        self._loop = loop
        self._connection_lost_callback = connection_lost_callback

        self._buffer = asyncio.Queue(loop=self._loop)
        self._recv_queue = deque([])
        self._send_queue = asyncio.Queue(loop=self._loop)
        self._acknak_queue = asyncio.Queue(loop=self._loop)
        self._next_all_link_rec_nak_retries = 0
        self._aldb_devices = {}
        self._devices = LinkedDevices(loop, workdir)
        self._poll_devices = poll_devices
        self._load_aldb = load_aldb
        self._write_transport_lock = asyncio.Lock(loop=self._loop)
        self._message_callbacks = MessageCallback()
        self._x10_address = None

        # Callback lists
        self._cb_load_all_link_db_done = []
        self._cb_device_not_active = []

        super().__init__(self, "000000", 0x03, None, None, "", "")

        self.transport = None

        self._register_message_handlers()
        self._writer_task = None
        self._restart_writer = False
        self.restart_writing()

    # public properties
    @property
    def devices(self):
        """Return the list of devices linked to the IM."""
        return self._devices

    @property
    def loop(self):
        """Return the asyncio loop."""
        return self._loop

    @property
    def message_callbacks(self):
        """Return the list of message callbacks."""
        return self._message_callbacks

    # asyncio.protocol interface methods
    def connection_made(self, transport):
        """Complete the network connection.

        Called when asyncio.Protocol establishes the network connection.
        """
        raise NotImplementedError

    def data_received(self, data):
        """Receive data from the protocol.

        Called when asyncio.Protocol detects received data from network.
        """
        _LOGGER.debug("Starting: data_received")
        _LOGGER.debug(
            "Received %d bytes from PLM: %s", len(data), binascii.hexlify(data)
        )
        self._buffer.put_nowait(data)
        asyncio.ensure_future(self._peel_messages_from_buffer(), loop=self._loop)

        _LOGGER.debug("Finishing: data_received")

    def connection_lost(self, exc):
        """Reestablish the connection to the transport.

        Called when asyncio.Protocol loses the network connection.
        """
        if exc is None:
            _LOGGER.warning("End of file received from Insteon Modem")
        else:
            _LOGGER.warning("Lost connection to Insteon Modem: %s", exc)

        self.transport = None
        asyncio.ensure_future(self.pause_writing(), loop=self.loop)

        if self._connection_lost_callback:
            self._connection_lost_callback()

    # Methods used to trigger callbacks for specific events
    def add_device_callback(self, callback):
        """Register a callback for when a matching new device is seen."""
        self.devices.add_device_callback(callback)

    def add_all_link_done_callback(self, callback):
        """Register a callback to be invoked when the ALDB is loaded."""
        _LOGGER.debug("Added new callback %s ", callback)
        self._cb_load_all_link_db_done.append(callback)

    def add_device_not_active_callback(self, callback):
        """Register callback to be invoked when a device is not responding."""
        _LOGGER.debug("Added new callback %s ", callback)
        self._cb_device_not_active.append(callback)

    # Public methods
    def poll_devices(self):
        """Request status updates from each device."""
        for addr in self.devices:
            device = self.devices[addr]
            if not device.address.is_x10:
                device.async_refresh_state()

    def send_msg(self, msg, wait_nak=True, wait_timeout=WAIT_TIMEOUT):
        """Place a message on the send queue for sending.

        Message are sent in the order they are placed in the queue.
        """
        msg_info = MessageInfo(msg=msg, wait_nak=wait_nak, wait_timeout=wait_timeout)
        _LOGGER.debug("Queueing msg: %s", msg)
        self._send_queue.put_nowait(msg_info)

    def start_all_linking(self, mode, group):
        """Put the IM into All-Linking mode.

        Puts the IM into All-Linking mode for 4 minutes.

        Parameters:
            mode: 0 | 1 | 3 | 255
                  0 - PLM is responder
                  1 - PLM is controller
                  3 - Device that initiated All-Linking is Controller
                255 = Delete All-Link
            group: All-Link group number (0 - 255)

        """
        msg = StartAllLinking(mode, group)
        self.send_msg(msg)

    def add_x10_device(self, housecode, unitcode, feature="OnOff"):
        """Add an X10 device based on a feature description.

        Current features are:
        - OnOff
        - Dimmable
        - Sensor
        - AllUnitsOff
        - AllLightsOn
        - AllLightsOff
        """
        device = insteonplm.devices.create_x10(self, housecode, unitcode, feature)
        if device:
            self.devices[device.address.id] = device
        return device

    def monitor_mode(self):
        """Put the Insteon Modem in monitor mode."""
        msg = SetIMConfiguration(0x40)
        self.send_msg(msg)

    def device_not_active(self, addr):
        """Handle inactive devices."""
        self.aldb_device_handled(addr)
        for callback in self._cb_device_not_active:
            callback(addr)

    def aldb_device_handled(self, addr):
        """Remove device from ALDB device list."""
        if isinstance(addr, Address):
            remove_addr = addr.id
        else:
            remove_addr = addr
        try:
            self._aldb_devices.pop(remove_addr)
            _LOGGER.debug("Removed ALDB device %s", remove_addr)
        except KeyError:
            _LOGGER.debug("Device %s not in ALDB device list", remove_addr)

        _LOGGER.debug("Entries remaining in ALDB device discovery list: %d",
                      len(self._aldb_devices))

    def manage_aldb_record(
        self, control_code, control_flags, group, address, data1, data2, data3
    ):
        """Update an IM All-Link record.

        Control Code values:
        - 0x00 Find First Starting at the top of the ALDB, search for the
        first ALL-Link Record matching the <ALL-Link Group> and <ID> in bytes
        5 – 8. The search ignores byte 4, <ALL-Link Record Flags>. You will
        receive an ACK at the end of the returned message if such an ALL-Link
        Record exists, or else a NAK if it doesn’t.  If the record exists, the
        IM will return it in an ALL-Link Record Response (0x51) message.

        - 0x01 Find Next Search for the next ALL-Link Record following the one
        found using <Control Code> 0x00 above.  This allows you to find both
        Controller and Responder records for a given <ALL-Link Group> and
        <ID>. Be sure to use the same <ALL-Link Group> and <ID> (bytes 5 – 8)
        as you used for <Control Code> 0x00. You will receive an ACK at the
        end of the returned message if another matching ALL-Link Record
        exists, or else a NAK if it doesn’t.  If the record exists, the IM
        will return it in an ALL-Link Record Response (0x51) message.

        - 0x20 Modify First Found or Add Modify an existing or else add a new
        ALL-Link Record for either a Controller or Responder. Starting at the
        top of the ALDB, search for the first ALL-Link Record matching the
        <ALL-Link Group> and <ID> in bytes 5 – 8.  The search ignores byte 4,
        <ALL-Link Record Flags>. If such an ALL-Link Record exists, overwrite
        it with the data in bytes 4 – 11; otherwise, create a new ALL-Link
        Record using bytes 4 – 11. Note that the IM will copy
        <ALL-Link Record Flags> you supplied in byte 4 below directly into the
        <ALL-Link Record Flags> byte of the ALL-Link Record in an ALDB-L
        (linear) database.  Use caution, because you can damage an ALDB-L if
        you misuse this Command.  For instance, if you zero the
        <ALL-Link Record Flags> byte in the first ALL-Link Record, the IM’s
        ALDB-L database will then appear empty.

        - 0x40 Modify First Controller Found or Add Modify an existing or else
        add a new Controller (master) ALL-Link Record. Starting at the top of
        the ALDB, search for the first ALL-Link Controller Record matching the
        <ALL-Link Group> and <ID> in bytes 5 – 8.  An ALL-Link Controller
        Record has bit 6 of its <ALL-Link Record Flags> byte set to 1. If such
        a Controller ALL-Link Record exists, overwrite it with the data in
        bytes 5 – 11; otherwise, create a new ALL-Link Record using bytes
        5 – 11.  In either case, the IM will set bit 6 of the
        <ALL-Link Record Flags> byte in the ALL-Link Record to 1 to indicate
        that the record is for a Controller.

        - 0x41 Modify First Responder Found or Add Modify an existing or else
        add a new Responder (slave) ALLLink Record. Starting at the top of the
        ALDB, search for the first ALL-Link Responder Record matching the
        <ALL-Link Group> and <ID> in bytes 5 – 8.  An ALL-Link Responder
        Record has bit 6 of its <ALL-Link Record Flags> byte cleared to 0. If
        such a Responder ALL-Link Record exists, overwrite it with the data in
        bytes 5 – 11; otherwise, create a new ALL-Link Record using bytes
        5 – 11.  In either case, The IM will clear bit 6 of the
        <ALL-Link Record Flags> byte in the ALL-Link Record to 0 to indicate
        that the record is for a Responder.

        - 0x80 Delete First Found Delete an ALL-Link Record. Starting at the
        top of the ALDB, search for the first ALL-Link Record matching the
        <ALL-Link Group> and <ID> in bytes 5 – 8.  The search ignores byte 4,
        <ALL-Link Record Flags>. You will receive an ACK at the end of the
        returned message if such an ALL-Link Record existed and was deleted,
        or else a NAK no such record exists.
        """
        msg = ManageAllLinkRecord(
            control_code, control_flags, group, address, data1, data2, data3
        )
        self.send_msg(msg)

    async def pause_writing(self):
        """Pause writing."""
        self._restart_writer = False
        if self._writer_task:
            self._send_queue.put_nowait(None)
        await asyncio.sleep(0.1)

    # pylint: disable=unused-argument
    def restart_writing(self, task=None):
        """Resume writing."""
        if self._restart_writer and not self._write_transport_lock.locked():
            self._writer_task = asyncio.ensure_future(
                self._get_message_from_send_queue(), loop=self._loop
            )
            self._writer_task.add_done_callback(self.restart_writing)

    async def close(self):
        """Close all writers for all devices for a clean shutdown."""
        await self.pause_writing()
        await asyncio.sleep(0, loop=self._loop)

    def trigger_group_on(self, group):
        """Trigger an All-Link Group on."""
        from .messages.standardSend import StandardSend
        from .constants import COMMAND_LIGHT_ON_0X11_NONE

        target = Address(bytearray([0x00, 0x00, group]))
        flags = 0xCF
        msg = StandardSend(target, COMMAND_LIGHT_ON_0X11_NONE, cmd2=0xFF, flags=flags)
        self.send_msg(msg)
        dev_list = self._find_scene(group)
        _LOGGER.debug("Scene %d turned on", group)
        for addr in dev_list:
            device = self._devices[addr.id]
            flags = 0x4F
            msg = StandardSend(
                device.address, COMMAND_LIGHT_ON_0X11_NONE, cmd2=group, flags=flags
            )
            self.send_msg(msg)
            if hasattr(device, "async_refresh_state"):
                _LOGGER.debug("Checking status of device %s", addr.human)
                device.async_refresh_state()

    def trigger_group_off(self, group):
        """Trigger an All-Link Group off."""
        from .messages.standardSend import StandardSend
        from .constants import COMMAND_LIGHT_OFF_0X13_0X00

        target = Address(bytearray([0x00, 0x00, group]))
        flags = 0xCF
        msg = StandardSend(target, COMMAND_LIGHT_OFF_0X13_0X00, flags=flags)
        self.send_msg(msg)
        dev_list = self._find_scene(group)
        _LOGGER.debug("Scene %d turned off", group)
        for addr in dev_list:
            device = self._devices[addr.id]
            flags = 0x4F
            msg = StandardSend(
                device.address, COMMAND_LIGHT_OFF_0X13_0X00, cmd2=group, flags=flags
            )
            self.send_msg(msg)
            if hasattr(device, "async_refresh_state"):
                _LOGGER.debug("Checking status of device %s", addr.human)
                device.async_refresh_state()

    def _find_scene(self, group):
        """Identify all devices that are part of a scene."""
        device_list = []
        for rec_num in self._aldb:
            rec = self._aldb[rec_num]
            _LOGGER.debug("Checking record for scene: %s", rec)
            if rec.control_flags.is_controller and rec.group == group:
                if rec.address not in device_list:
                    device_list.append(rec.address)
        for addr in self._devices:
            device = self._devices[addr]
            aldb = device.aldb
            for mem_addr in aldb:
                rec = aldb[mem_addr]
                _LOGGER.debug("Checking record for scene: %s", rec)
                if (
                    rec.control_flags.is_in_use
                    and rec.group == group
                    and rec.address == self._address
                ):
                    if rec.address not in device_list:
                        device_list.append(device.address)
        return device_list

    async def _setup_devices(self):
        await self.devices.load_saved_device_info()

        _LOGGER.debug("Found %d saved devices", len(self.devices.saved_devices))
        self._get_plm_info()
        self.devices.add_known_devices(self)

        # Comment out the following lines for testing
        if self._load_aldb:
            self._load_all_link_database()
        else:
            self._complete_setup()

        _LOGGER.debug("Ending _setup_devices in IM")

    # pylint: disable=broad-except
    async def _get_message_from_send_queue(self):
        _LOGGER.debug("Starting Insteon Modem write message from send queue")
        _LOGGER.debug("Aquiring write lock")
        await self._write_transport_lock.acquire()
        while self._restart_writer:
            # wait for an item from the queue
            msg_info = await self._send_queue.get()
            if msg_info is None:
                self._restart_writer = False
                return
            message_sent = False
            try:
                while not message_sent:
                    message_sent = await self._write_message(msg_info)
                await asyncio.sleep(msg_info.wait_timeout, loop=self._loop)
            except asyncio.CancelledError:
                _LOGGER.info("Stopping Insteon Modem writer due to " "CancelledError")
                self._restart_writer = False
            except GeneratorExit:
                _LOGGER.error("Stopping Insteon Modem writer due to " "GeneratorExit")
                self._restart_writer = False
            except Exception as e:
                _LOGGER.error("Restarting Insteon Modem writer due to %s", str(e))
                _LOGGER.error("MSG: %s", str(msg_info.msg))
                self._restart_writer = True
        if self._write_transport_lock.locked():
            self._write_transport_lock.release()
        _LOGGER.debug("Ending Insteon Modem write message from send queue")

    def _get_plm_info(self):
        """Request PLM Info."""
        _LOGGER.info("Requesting Insteon Modem Info")
        msg = GetImInfo()
        self.send_msg(msg, wait_nak=True, wait_timeout=0.5)

    async def _write_message(self, msg_info: MessageInfo):
        _LOGGER.debug("TX: %s:%s", id(msg_info.msg), msg_info.msg)
        is_sent = False
        if not self.transport.is_closing():
            self.transport.write(msg_info.msg.bytes)
            if msg_info.wait_nak:
                _LOGGER.debug("Waiting for ACK or NAK message")
                is_sent = await self._wait_ack_nak(msg_info.msg)
            else:
                is_sent = True
        else:
            _LOGGER.debug("Transport is not open, waiting 5 seconds")
            is_sent = False
            await asyncio.sleep(5, loop=self._loop)
        return is_sent

    async def _wait_ack_nak(self, msg):
        is_sent = False
        is_ack_nak = False
        try:
            with async_timeout.timeout(ACKNAK_TIMEOUT):
                while not is_ack_nak:
                    acknak = await self._acknak_queue.get()
                    _LOGGER.debug('Processing _acknak_queue acknak: %s '
                                  'msg: %s', id(acknak), id(msg))
                    is_ack_nak = self._msg_is_ack_nak(msg, acknak)
                    is_sent = self._msg_is_sent(acknak)
        except asyncio.TimeoutError:
            _LOGGER.debug("No ACK or NAK message received.")
            is_sent = False
        return is_sent

    # pylint: disable=no-self-use
    def _msg_is_ack_nak(self, msg, acknak):
        if not hasattr(acknak, "isack"):
            return False
        if msg.matches_pattern(acknak):
            _LOGGER.debug("ACK or NAK received")
            return True
        return False

    # pylint: disable=no-self-use
    def _msg_is_sent(self, acknak):
        # All Link record NAK is a valid last record response
        # However, we want to retry 3 times to make sure
        # it is a valid last record response and not a true NAK
        if (
            acknak.code == GetFirstAllLinkRecord.code
            or acknak.code == GetNextAllLinkRecord.code
        ) and acknak.isnak:
            return True

        return acknak.isack

    def _load_all_link_database(self):
        """Load the ALL-Link Database into object."""
        _LOGGER.debug("Starting: _load_all_link_database")
        self.devices.state = "loading"
        self._get_first_all_link_record()
        _LOGGER.debug("Ending: _load_all_link_database")

    def _get_first_all_link_record(self):
        """Request first ALL-Link record."""
        _LOGGER.debug("Starting: _get_first_all_link_record")
        _LOGGER.info("Requesting ALL-Link Records")
        if self.aldb.status == ALDBStatus.LOADED:
            self._next_all_link_rec_nak_retries = 3
            self._handle_get_next_all_link_record_nak(None)
            return
        self.aldb.clear()
        self._next_all_link_rec_nak_retries = 0
        msg = GetFirstAllLinkRecord()
        self.send_msg(msg, wait_nak=True, wait_timeout=0.5)
        _LOGGER.debug("Ending: _get_first_all_link_record")

    def _get_next_all_link_record(self):
        """Request next ALL-Link record."""
        _LOGGER.debug("Starting: _get_next_all_link_record")
        _LOGGER.debug("Requesting Next All-Link Record")
        msg = GetNextAllLinkRecord()
        self.send_msg(msg, wait_nak=True, wait_timeout=0.5)
        _LOGGER.debug("Ending: _get_next_all_link_record")

    def _new_device_added(self, device):
        self.aldb_device_handled(device.address.id)
        if self._poll_devices:
            device.async_refresh_state()

    # Inbound message handlers sepcific to the IM
    def _register_message_handlers(self):
        template_all_link_response = AllLinkRecordResponse(
            None, None, None, None, None, None
        )
        template_get_im_info = GetImInfo()
        template_next_all_link_rec = GetNextAllLinkRecord(acknak=MESSAGE_NAK)
        template_x10_send = X10Send(None, None, MESSAGE_ACK)
        template_x10_received = X10Received(None, None)

        # self._message_callbacks.add(
        #    template_assign_all_link,
        #    self._handle_assign_to_all_link_group)

        self._message_callbacks.add(
            template_all_link_response, self._handle_all_link_record_response
        )

        self._message_callbacks.add(template_get_im_info, self._handle_get_plm_info)

        self._message_callbacks.add(
            template_next_all_link_rec, self._handle_get_next_all_link_record_nak
        )

        self._message_callbacks.add(template_x10_send, self._handle_x10_send_receive)

        self._message_callbacks.add(
            template_x10_received, self._handle_x10_send_receive
        )

    async def _peel_messages_from_buffer(self):
        lastlooplen = 0
        worktodo = True
        buffer = bytearray()
        while worktodo:
            buffer.extend(self._unpack_buffer())
            if len(buffer) < 2:
                worktodo = False
                break
            _LOGGER.debug("Total buffer: %s", binascii.hexlify(buffer))
            msg, buffer = insteonplm.messages.create(buffer)

            if msg is not None:
                # _LOGGER.debug('Msg buffer: %s', msg.hex)
                self._recv_queue.appendleft(msg)

            # _LOGGER.debug('Post buffer: %s', binascii.hexlify(buffer))
            if len(buffer) < 2:
                _LOGGER.debug("Buffer too short to have a message")
                worktodo = False
                break

            if len(buffer) == lastlooplen:
                _LOGGER.debug("Buffer size did not change wait for more data")
                worktodo = False
                break

            lastlooplen = len(buffer)
        if buffer:
            buffer.extend(self._unpack_buffer())
            self._buffer.put_nowait(buffer)

        _LOGGER.debug("Messages in queue: %d", len(self._recv_queue))
        worktodo = True
        while worktodo:
            try:
                self._process_recv_queue()
            except IndexError:
                _LOGGER.debug("Last item in self._recv_queue reached.")
                worktodo = False

    def _process_recv_queue(self):
        msg = self._recv_queue.pop()

        _LOGGER.debug("RX: %s:%s", id(msg), msg)
        callbacks = self._message_callbacks.get_callbacks_from_message(msg)
        if hasattr(msg, "isack") or hasattr(msg, "isnak"):
            self._acknak_queue.put_nowait(msg)
        if hasattr(msg, "address"):
            device = self.devices[msg.address.hex]
            if device:
                device.receive_message(msg)
            else:
                try:
                    device = self._aldb_devices[msg.address.id]
                    if device:
                        device.receive_message(msg)
                except KeyError:
                    pass
        for callback in callbacks:
            self._loop.call_soon(callback, msg)

    def _unpack_buffer(self):
        buffer = bytearray()
        while not self._buffer.empty():
            buffer.extend(self._buffer.get_nowait())
        return buffer

    def _refresh_aldb(self):
        self.aldb.clear()
        self._load_all_link_database()

    def _handle_all_link_record_response(self, msg):

        _LOGGER.debug("Found all link %s record for group 0x%02x, device %s",
                      "control" if msg.isController else "respond",
                      msg.group, msg.address.human)
        cat = msg.linkdata1
        subcat = msg.linkdata2
        product_key = msg.linkdata3
        rec_num = len(self._aldb)
        self._aldb[rec_num] = ALDBRecord(
            rec_num, msg.controlFlags, msg.group, msg.address, cat, subcat, product_key
        )
        if self.devices[msg.address.id] is None:
            _LOGGER.debug(
                "ALDB Data: address %s data1: %02x " "data1: %02x data3: %02x",
                msg.address.hex,
                cat,
                subcat,
                product_key,
            )

            # Get a device from the ALDB based on cat, subcat and product_key
            device = self.devices.create_device_from_category(
                self, msg.address, cat, subcat, product_key
            )

            # If a device is returned and that device is of a type that
            # can not be queried for the product data, then save the device
            # with the data from the IPDB
            if device is not None:
                if (
                    device.prod_data_in_aldb
                    or self.devices.has_override(device.address.id)
                    or self.devices.has_saved(device.address.id)
                ):
                    if self.devices[device.id] is None:
                        self.devices[device.id] = device
                        _LOGGER.debug(
                            "Device with id %s added to device list from ALDB data.",
                            device.id,
                        )

        # Check again that the device is not already added, otherwise queue it
        # up for Get ID request
        if not self.devices[msg.address.id]:
            if msg.address.id in self._aldb_devices:
                _LOGGER.debug("Device %s already queued for Get ID request",
                              msg.address.hex)
            else:
                _LOGGER.info("Found new device %s. Queueing device for "
                             "Get ID request", msg.address.id)
                unknowndevice = self.devices.create_device_from_category(
                    self, msg.address.hex, None, None, None)
                self._aldb_devices[msg.address.id] = unknowndevice
        else:
            _LOGGER.debug("Device %s already loaded from saved data or "
                          "overrides", msg.address.hex)

        self._next_all_link_rec_nak_retries = 0
        self._get_next_all_link_record()

    def _handle_get_next_all_link_record_nak(self, msg):
        # When the last All-Link record is reached the PLM sends a NAK
        if self._next_all_link_rec_nak_retries < 3:
            self._next_all_link_rec_nak_retries += 1
            self._get_next_all_link_record()
            return
        self._aldb.status = ALDBStatus.LOADED
        _LOGGER.debug(
            "All-Link device records found in ALDB: %d", len(self._aldb_devices)
        )

        while self._cb_load_all_link_db_done:
            callback = self._cb_load_all_link_db_done.pop()
            callback()

        self._get_device_info()

    def _get_device_info(self):
        _LOGGER.debug("Starting _get_device_info")
        # Remove saved records for devices found in the ALDB
        for addr in self.devices:
            self.aldb_device_handled(addr)

        self._complete_setup()

        self.devices.add_device_callback(self._new_device_added)

        for addr in self._aldb_devices:
            _LOGGER.debug("Getting device info for %s", Address(addr).human)
            self._aldb_devices[addr].id_request()

        _LOGGER.debug("Ending _get_device_info")

    def _complete_setup(self):
        self.devices.save_device_info()
        if self._poll_devices:
            self._loop.call_soon(self.poll_devices)

    def _handle_nak(self, msg_info):
        if (
            msg_info.msg.code == GetFirstAllLinkRecord.code
            or msg_info.msg.code == GetNextAllLinkRecord.code
        ):
            return
        _LOGGER.debug("No response or NAK message received")
        self.send_msg(msg_info.msg, msg_info.wait_nak, msg_info.wait_timeout)

    def _handle_get_plm_info(self, msg):
        from insteonplm.devices import ALDB

        _LOGGER.debug("Starting _handle_get_plm_info")
        from insteonplm.devices.ipdb import IPDB

        ipdb = IPDB()
        self._address = msg.address
        self._cat = msg.category
        self._subcat = msg.subcategory
        self._product_key = msg.firmware
        product = ipdb[[self._cat, self._subcat]]
        self._description = product.description
        self._model = product.model
        self._aldb = ALDB(self._send_msg, self._plm.loop, self._address)

        _LOGGER.debug("Ending _handle_get_plm_info")

    # X10 Device methods
    def x10_all_units_off(self, housecode):
        """Send the X10 All Units Off command."""
        if isinstance(housecode, str):
            housecode = housecode.upper()
        else:
            raise TypeError("Housecode must be a string")
        msg = X10Send.command_msg(housecode, X10_COMMAND_ALL_UNITS_OFF)
        self.send_msg(msg)
        self._x10_command_to_device(housecode, X10_COMMAND_ALL_UNITS_OFF, msg)

    def x10_all_lights_off(self, housecode):
        """Send the X10 All Lights Off command."""
        msg = X10Send.command_msg(housecode, X10_COMMAND_ALL_LIGHTS_OFF)
        self.send_msg(msg)
        self._x10_command_to_device(housecode, X10_COMMAND_ALL_LIGHTS_OFF, msg)

    def x10_all_lights_on(self, housecode):
        """Send the X10 All Lights Off command."""
        msg = X10Send.command_msg(housecode, X10_COMMAND_ALL_LIGHTS_ON)
        self.send_msg(msg)
        self._x10_command_to_device(housecode, X10_COMMAND_ALL_LIGHTS_ON, msg)

    def _handle_x10_send_receive(self, msg):
        housecode_byte, unit_command_byte = rawX10_to_bytes(msg.rawX10)
        housecode = byte_to_housecode(housecode_byte)
        if msg.flag == 0x00:
            unitcode = byte_to_unitcode(unit_command_byte)
            self._x10_address = Address.x10(housecode, unitcode)
            if self._x10_address:
                device = self.devices[self._x10_address.id]
                if device:
                    device.receive_message(msg)
        else:
            self._x10_command_to_device(housecode, unit_command_byte, msg)

    def _x10_command_to_device(self, housecode, command, msg):
        if isinstance(housecode, str):
            housecode = housecode.upper()
        else:
            raise TypeError("Housecode must be a string")
        if x10_command_type(command) == X10CommandType.DIRECT:
            if self._x10_address and self.devices[self._x10_address.id]:
                if self._x10_address.x10_housecode == housecode:
                    self.devices[self._x10_address.id].receive_message(msg)
        else:
            for addr in self.devices:
                if self.devices[addr].address.is_x10:
                    if self.devices[addr].address.x10_housecode == housecode:
                        self.devices[addr].receive_message(msg)
        self._x10_address = None


class PLM(IM):
    """Provide an Insteon PowerLinc Modem device.

    This class is expected to be wrapped inside a Connection class object
    which will maintain the socket and handle auto-reconnects.

        :param connection_lost_callback:
            called when connection is lost to device (optional)
        :param loop:
            asyncio event loop (optional)
        :param workdir:
            Working directory name to save device information (optional)

        :type: connection_lost_callback:
            callable
        :type loop:
            asyncio.loop
        :type workdir:
            string - valid directory path
    """

    # asyncio.protocol interface methods
    def connection_made(self, transport):
        """Start the PLM connection process.

        Called when asyncio.Protocol establishes the network connection.
        """
        _LOGGER.info("Connection established to PLM")
        self.transport = transport

        self._restart_writer = True
        self.restart_writing()

        # Testing to see if this fixes the 2413S issue
        self.transport.serial.timeout = 1
        self.transport.serial.write_timeout = 1
        self.transport.set_write_buffer_limits(128)
        # limit = self.transport.get_write_buffer_size()
        # _LOGGER.debug('Write buffer size is %d', limit)
        if self._aldb.status != ALDBStatus.LOADED:
            asyncio.ensure_future(self._setup_devices(), loop=self._loop)


class Hub(IM):
    """Provide an Insteon Hub device.

    This class is expected to be wrapped inside a Connection class object
    which will maintain the socket and handle auto-reconnects.

        :param connection_lost_callback:
            called when connection is lost to device (optional)
        :param loop:
            asyncio event loop (optional)
        :param workdir:
            Working directory name to save device information (optional)

        :type: connection_lost_callback:
            callable
        :type loop:
            asyncio.loop
        :type workdir:
            string - valid directory path
    """

    # asyncio.protocol interface methods
    def connection_made(self, transport):
        """Start the Hub connection process.

        Called when asyncio.Protocol establishes the network connection.
        """
        _LOGGER.info("Connection established to Hub")
        _LOGGER.debug("Transport: %s", transport)
        self.transport = transport

        self._restart_writer = True
        self.restart_writing()

        if self._aldb.status != ALDBStatus.LOADED:
            asyncio.ensure_future(self._setup_devices(), loop=self._loop)



"""Mock PLM class for testing devices."""
import logging
from insteonplm.messagecallback import MessageCallback
from insteonplm.linkedDevices import LinkedDevices

_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.DEBUG)


class MockPLM():
    """Mock PLM class for testing devices."""

    def __init__(self, loop=None):
        """Init the MockPLM class."""
        self.sentmessage = ''
        self._message_callbacks = MessageCallback()
        self.loop = loop
        self.devices = LinkedDevices()

    @property
    def message_callbacks(self):
        """Return the message callback list."""
        return self._message_callbacks

    # pylint: disable=unused-argument
    def send_msg(self, msg, wait_nak=True, wait_timeout=2):
        """Send a message mock routine."""
        _LOGGER.debug('TX: %s:%s', id(msg), msg)
        self.sentmessage = msg.hex

    def message_received(self, msg):
        """Fake a message being received by the PLM."""
        _LOGGER.debug('RX: %s:%s', id(msg), msg)
        if hasattr(msg, 'address'):
            device = self.devices[msg.address.id]
            if device:
                device.receive_message(msg)
            else:
                _LOGGER.info('Received message for unknown device %s',
                             msg.address)
        for callback in (
                self._message_callbacks.get_callbacks_from_message(msg)):
            callback(msg)

    # pylint: disable=unused-argument
    def start_all_linking(self, linkcode, group):
        """Fake start all linking."""
        self.sentmessage = b'02112233445566'



"""INSTEON Message Button Event Report."""
from insteonplm.messages.message import Message
from insteonplm.constants import (
    MESSAGE_BUTTON_EVENT_REPORT_0X54,
    MESSAGE_BUTTON_EVENT_REPORT_SIZE,
)


class ButtonEventReport(Message):
    """Insteon Button Event Report Message Received.

    Message type 0x54
    """

    _code = MESSAGE_BUTTON_EVENT_REPORT_0X54
    _sendSize = MESSAGE_BUTTON_EVENT_REPORT_SIZE
    _receivedSize = MESSAGE_BUTTON_EVENT_REPORT_SIZE
    _description = "INSTEON Standard Message Received"

    def __init__(self, event):
        """Init the ButtonEventReport Class."""
        self._event = event

        self._events = {
            0x02: "SET button tapped",
            0x03: "SET button press and hold",
            0x04: "SET button released",
            0x12: "Button 2 tapped",
            0x13: "Button 2 press and hold",
            0x14: "Button 2 released",
            0x22: "Button 3 tapped",
            0x23: "Button 3 press and hold",
            0x24: "Button 3 released",
        }

    @classmethod
    def from_raw_message(cls, rawmessage):
        """Create message from raw byte stream."""
        return ButtonEventReport(rawmessage[2])

    @property
    def event(self):
        """Return the button event property."""
        return self._event

    @property
    def eventText(self):
        """Return the button event text."""
        return self._events.get(self.event, None)

    def _message_properties(self):
        return [{"event": self._event}]



"""INSTEON Extended Receive Message Type 0x51."""

from insteonplm.constants import (
    MESSAGE_EXTENDED_MESSAGE_RECEIVED_0X51,
    MESSAGE_EXTENDED_MESSAGE_RECEIVED_SIZE,
)
from insteonplm.address import Address
from insteonplm.messages.message import Message
from insteonplm.messages.messageFlags import MessageFlags
from insteonplm.messages.userdata import Userdata


class ExtendedReceive(Message):
    """Insteon Extended Length Message Received.

    Message type 0x51
    """

    _code = MESSAGE_EXTENDED_MESSAGE_RECEIVED_0X51
    _sendSize = MESSAGE_EXTENDED_MESSAGE_RECEIVED_SIZE
    _receivedSize = MESSAGE_EXTENDED_MESSAGE_RECEIVED_SIZE
    _description = "INSTEON Extended Message Received"

    def __init__(self, address, target, commandtuple, userdata, cmd2=None, flags=0x10):
        """Init the ExtendedRecieve message class."""
        if commandtuple.get("cmd1", None) is not None:
            cmd1 = commandtuple["cmd1"]
            cmd2out = commandtuple["cmd2"]
        else:
            raise ValueError

        if cmd2 is not None:
            cmd2out = cmd2

        if cmd2out is None:
            raise ValueError

        self._address = Address(address)
        self._target = Address(target)
        self._messageFlags = MessageFlags(flags)
        # self._messageFlags.extended = 1
        self._cmd1 = cmd1
        self._cmd2 = cmd2out
        self._userdata = Userdata(userdata)

    @classmethod
    def from_raw_message(cls, rawmessage):
        """Create message from raw byte stream."""
        userdata = Userdata.from_raw_message(rawmessage[11:25])
        return ExtendedReceive(
            rawmessage[2:5],
            rawmessage[5:8],
            {"cmd1": rawmessage[9], "cmd2": rawmessage[10]},
            userdata,
            flags=rawmessage[8],
        )

    # pylint: disable=protected-access
    @classmethod
    def template(
        cls,
        address=None,
        target=None,
        commandtuple=None,
        userdata=None,
        cmd2=-1,
        flags=None,
    ):
        """Create message template for callbacks."""
        msgraw = bytearray([0x02, cls._code])
        msgraw.extend(bytes(cls._receivedSize))
        msg = ExtendedReceive.from_raw_message(msgraw)

        if commandtuple:
            cmd1 = commandtuple.get("cmd1")
            cmd2out = commandtuple.get("cmd2")
        else:
            cmd1 = None
            cmd2out = None

        if cmd2 is not -1:
            cmd2out = cmd2

        msg._address = Address(address)
        msg._target = Address(target)
        msg._messageFlags = MessageFlags(flags)
        msg._cmd1 = cmd1
        msg._cmd2 = cmd2out
        msg._userdata = Userdata.create_pattern(userdata)
        return msg

    @property
    def address(self):
        """Return address of sending device."""
        return self._address

    @property
    def target(self):
        """Return address of target device."""
        return self._target

    @property
    def cmd1(self):
        """Return cmd1 property of message."""
        return self._cmd1

    @property
    def cmd2(self):
        """Return cmd2 property of message."""
        return self._cmd2

    @property
    def userdata(self):
        """Return user data of extended message."""
        return self._userdata

    @property
    def flags(self):
        """Return message flags."""
        return self._messageFlags

    @property
    def targetLow(self):
        """Return the low byte of the target address field.

        Used in All-Link Cleanup messages.
        """
        low_byte = None
        if self.target is not None and self._messageFlags.isBroadcast:
            low_byte = self.target.bytes[0]
        return low_byte

    @property
    def targetMed(self):
        """Return the middle byte of the target address field.

        Used in All-Link Cleanup messages.
        """
        med_byte = None
        if self.target is not None and self._messageFlags.isBroadcast:
            med_byte = self.target.bytes[1]
        return med_byte

    @property
    def targetHi(self):
        """Return the high byte of the target address field.

        Used in All-Link Cleanup messages.
        """
        hi_byte = None
        if self.target is not None and self._messageFlags.isBroadcast:
            hi_byte = self.target.bytes[2]
        return hi_byte

    def _message_properties(self):
        return [
            {"address": self._address},
            {"target": self._target},
            {"flags": self._messageFlags},
            {"cmd1": self._cmd1},
            {"cmd2": self._cmd2},
            {"userdata": self._userdata},
        ]



"""INSTEON Message All-Link Complete."""

from insteonplm.address import Address
from insteonplm.constants import (
    MESSAGE_ALL_LINKING_COMPLETED_0X53,
    MESSAGE_ALL_LINKING_COMPLETED_SIZE,
)
from insteonplm.messages.message import Message


class AllLinkComplete(Message):
    """INSTEON ALL-Linking Completed Message.

    Message type 0x53
    """

    _code = MESSAGE_ALL_LINKING_COMPLETED_0X53
    _sendSize = MESSAGE_ALL_LINKING_COMPLETED_SIZE
    _receivedSize = MESSAGE_ALL_LINKING_COMPLETED_SIZE
    _description = "INSTEON ALL-Linking Completed Message Received"

    def __init__(self, linkcode, group, address, cat, subcat, firmware):
        """Init the AllLinkComplete Class."""
        self._linkcode = linkcode
        self._group = group
        self._address = Address(address)
        self._category = cat
        self._subcategory = subcat
        self._firmware = firmware

    @classmethod
    def from_raw_message(cls, rawmessage):
        """Create a message from a raw byte stream."""
        return AllLinkComplete(
            rawmessage[2],
            rawmessage[3],
            rawmessage[4:7],
            rawmessage[7],
            rawmessage[8],
            rawmessage[9],
        )

    @property
    def linkcode(self):
        """Return link record link code."""
        return self._linkcode

    @property
    def group(self):
        """Return link record group."""
        return self._group

    @property
    def address(self):
        """Return the device address."""
        return self._address

    @property
    def category(self):
        """Return the device category."""
        return self._category

    @property
    def subcategory(self):
        """Return the device subcategory."""
        return self._subcategory

    @property
    def firmware(self):
        """Return the device firmware version."""
        return self._firmware

    @property
    def isresponder(self):
        """Return if the link record is a responder."""
        return bool(self.linkcode == 0)

    @property
    def iscontroller(self):
        """Return if the link record is a controller."""
        return bool(self.linkcode == 1)

    @property
    def isdeleted(self):
        """Return if the link record is deleted."""
        return bool(self.linkcode == 0xFF)

    def _message_properties(self):
        return [
            {"linkcode": self._linkcode},
            {"group": self._group},
            {"address": self._address},
            {"category": self._category},
            {"subcategory": self._subcategory},
            {"firmware": self._firmware},
        ]



"""Sensor states."""
from enum import Enum
from insteonplm.constants import (
    COMMAND_LIGHT_ON_0X11_NONE,
    COMMAND_LIGHT_OFF_0X13_0X00,
    COMMAND_LIGHT_STATUS_REQUEST_0X19_0X01,
    MESSAGE_TYPE_BROADCAST_MESSAGE,
    MESSAGE_TYPE_ALL_LINK_CLEANUP,
    MESSAGE_TYPE_ALL_LINK_BROADCAST,
)
from insteonplm.messages.standardSend import StandardSend
from insteonplm.messages.standardReceive import StandardReceive
from insteonplm.messages.messageFlags import MessageFlags
from insteonplm.states import State


class LeakSensorState(Enum):
    """Enum to define dry/wet state of the leak sensor."""

    DRY = 0
    WET = 1


class SensorBase(State):
    """Base state representing a non-controlable variable value sensor.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
      register_updates(self, callback)
    """

    def __init__(
        self,
        address,
        statename,
        group,
        send_message_method,
        message_callbacks,
        defaultvalue=None,
    ):
        """Init the State state."""
        super().__init__(
            address,
            statename,
            group,
            send_message_method,
            message_callbacks,
            defaultvalue,
        )

        self._is_responder = False

        template_on_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_BROADCAST_MESSAGE, None),
            cmd2=None,
        )
        template_off_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_0X13_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_BROADCAST_MESSAGE, None),
            cmd2=None,
        )

        template_on_group = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )
        template_off_group = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_0X13_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )

        self._message_callbacks.add(
            template_on_broadcast, self._sensor_on_command_received
        )
        self._message_callbacks.add(
            template_off_broadcast, self._sensor_off_command_received
        )

        self._message_callbacks.add(template_on_group, self._sensor_on_command_received)
        self._message_callbacks.add(
            template_off_group, self._sensor_off_command_received
        )

    # pylint: disable=unused-argument
    def _sensor_on_command_received(self, msg):
        """Message handler for Standard or Extended sensor on messages.

        Message handler for Standard (0x50) or Extended (0x51) message commands
        0x11 Sensor On.  When a message is received any state listeners are
        updated with 0x11 for on.
        """
        self._update_subscribers(msg.cmd2)

    # pylint: disable=unused-argument
    def _sensor_off_command_received(self, msg):
        """Message handler for Standard or Extended sensor off messages.

        Message handler for Standard (0x50) or Extended (0x51) message commands
        0x11 Sensor On.  When a message is received any state listeners are
        updated with 0x13 for off.
        """
        self._update_subscribers(0x00)


class VariableSensor(SensorBase):
    """Device state representing a variable value sensor that is not controllable.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
      register_updates(self, callback)
    """


class OnOffSensor(SensorBase):
    """Device state representing an On/Off sensor that is not controllable.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
    - register_updates(self, callback)
    """

    # pylint: disable=unused-argument
    def _sensor_on_command_received(self, msg):
        """Message handler for Standard or Extended sensor on messages.

        Message handler for Standard (0x50) or Extended (0x51) message commands
        0x11 Sensor On.  When a message is received any state listeners are
        updated with 0x11 for on.
        """
        self._update_subscribers(0x01)


class SmokeCO2Sensor(SensorBase):
    """Device state representing a Smoke/CO2 sensor that is not controllable.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
    - register_updates(self, callback)
    """

    def __init__(
        self,
        address,
        statename,
        group,
        send_message_method,
        message_callbacks,
        defaultvalue=None,
    ):
        """Init the State state."""
        super().__init__(
            address,
            statename,
            group,
            send_message_method,
            message_callbacks,
            defaultvalue,
        )

        template_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_BROADCAST_MESSAGE, None),
        )
        template_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
        )
        template_group = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
        )

        self._message_callbacks.add(template_broadcast, self._sensor_state_received)
        self._message_callbacks.add(template_cleanup, self._sensor_state_received)
        self._message_callbacks.add(template_group, self._sensor_state_received)

    def _sensor_state_received(self, msg):
        self._update_subscribers(msg.targetHi)


class IoLincSensor(SensorBase):
    """Device state representing a I/O Linc Sensor that is not controllable.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
    - register_updates(self, callback)
    """

    def __init__(
        self,
        address,
        statename,
        group,
        send_message_method,
        message_callbacks,
        defaultvalue=None,
    ):
        """Init the IoLinkSensor."""
        super().__init__(
            address,
            statename,
            group,
            send_message_method,
            message_callbacks,
            defaultvalue,
        )

        self._updatemethod = self._send_status_request

        self._register_messages()

    def _register_messages(self):
        # The actual group number is 0x01 for the IOLinc sensor so this
        # is hard coded in these template messages
        template_close_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
            cmd2=0x01,
        )
        template_open_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_0X13_0X00,
            address=self._address,
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
            cmd2=0x01,
        )

        template_close_group = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, 0x01]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
        )
        template_open_group = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_OFF_0X13_0X00,
            address=self._address,
            target=bytearray([0x00, 0x00, 0x01]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
            cmd2=None,
        )

        self._message_callbacks.add(template_open_cleanup, self._open_message_received)
        self._message_callbacks.add(
            template_close_cleanup, self._close_message_received
        )

        self._message_callbacks.add(template_open_group, self._open_message_received)
        self._message_callbacks.add(template_close_group, self._close_message_received)

    def _send_status_request(self):
        status_command = StandardSend(
            self._address, COMMAND_LIGHT_STATUS_REQUEST_0X19_0X01
        )
        self._send_method(status_command, self._status_message_received)

    # pylint: disable=unused-argument
    def _open_message_received(self, msg):
        self._update_subscribers(0x00)

    # pylint: disable=unused-argument
    def _close_message_received(self, msg):
        self._update_subscribers(0x01)

    def _status_message_received(self, msg):
        if msg.cmd2 == 0x00:
            # 0x00 is open
            self._update_subscribers(0x00)
        else:
            self._update_subscribers(0x01)


class LeakSensorDryWet(State):
    """Water leak sensor for the Dry or Wet states.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
    - register_updates(self, callback)
    """

    def __init__(
        self,
        address,
        statename,
        group,
        send_message_method,
        message_callbacks,
        defaultvalue=None,
        dry_wet=None,
    ):
        """Init the LeakSensorDry state."""
        super().__init__(
            address,
            statename,
            group,
            send_message_method,
            message_callbacks,
            defaultvalue,
        )

        self._dry_wet_type = dry_wet
        self._dry_wet_callbacks = []

        template_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_BROADCAST_MESSAGE, None),
        )

        template_all_link = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
        )

        template_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
        )

        self._message_callbacks.add(template_broadcast, self._dry_wet_message_received)

        self._message_callbacks.add(template_all_link, self._dry_wet_message_received)

        self._message_callbacks.add(template_cleanup, self._dry_wet_message_received)

    def register_dry_wet_callback(self, callback):
        """Register the callback for the wet and dry state callbacks."""
        self._dry_wet_callbacks.append(callback)

    def set_value(self, dry_wet: LeakSensorState):
        """Set the value of the state to dry or wet."""
        value = 0
        if dry_wet == self._dry_wet_type:
            value = 1
        self._update_subscribers(value)

    # pylint: disable=unused-argument
    def _dry_wet_message_received(self, msg):
        """Report a dry or a wet state."""
        for callback in self._dry_wet_callbacks:
            callback(self._dry_wet_type)
        self._update_subscribers(0x01)


class LeakSensorHeartbeat(State):
    """Water leak sensor for the Dry state.

    Available properties are:
      value
      name
      address
      group

    Available methods are:
    - register_updates(self, callback)
    """

    def __init__(
        self,
        address,
        statename,
        group,
        send_message_method,
        message_callbacks,
        defaultvalue=None,
    ):
        """Init the LeakSensorDry state."""
        super().__init__(
            address,
            statename,
            group,
            send_message_method,
            message_callbacks,
            defaultvalue,
        )

        self._dry_wet_callbacks = []

        template_dry_broadcast = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            cmd2=self._group,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_BROADCAST_MESSAGE, None),
        )

        template_wet_broadcast = StandardReceive.template(
            commandtuple={"cmd1": 0x13, "cmd2": self._group},
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_BROADCAST_MESSAGE, None),
        )

        template_dry_all_link = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            cmd2=self._group,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
        )

        template_wet_all_link = StandardReceive.template(
            commandtuple={"cmd1": 0x13, "cmd2": self._group},
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_BROADCAST, None),
        )

        template_dry_cleanup = StandardReceive.template(
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
            cmd2=self._group,
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
        )

        template_wet_cleanup = StandardReceive.template(
            commandtuple={"cmd1": 0x13, "cmd2": self._group},
            address=self._address,
            target=bytearray([0x00, 0x00, self._group]),
            flags=MessageFlags.template(MESSAGE_TYPE_ALL_LINK_CLEANUP, None),
        )

        self._message_callbacks.add(template_dry_broadcast, self._dry_message_received)
        self._message_callbacks.add(template_wet_broadcast, self._wet_message_received)
        self._message_callbacks.add(template_dry_all_link, self._dry_message_received)
        self._message_callbacks.add(template_wet_all_link, self._wet_message_received)
        self._message_callbacks.add(template_dry_cleanup, self._dry_message_received)
        self._message_callbacks.add(template_wet_cleanup, self._wet_message_received)

    def register_dry_wet_callback(self, callback):
        """Register the callback for the wet and dry state callbacks."""
        self._dry_wet_callbacks.append(callback)

    def set_value(self, dry_wet: LeakSensorState):
        """Set the state to wet or dry."""
        if dry_wet == LeakSensorState.DRY:
            self._update_subscribers(0x11)
        else:
            self._update_subscribers(0x13)

    # pylint: disable=unused-argument
    def _dry_message_received(self, msg):
        """Report a dry state."""
        for callback in self._dry_wet_callbacks:
            callback(LeakSensorState.DRY)
        self._update_subscribers(0x11)

    # pylint: disable=unused-argument
    def _wet_message_received(self, msg):
        """Report a wet state."""
        for callback in self._dry_wet_callbacks:
            callback(LeakSensorState.WET)
        self._update_subscribers(0x13)



"""INSTEON Device Type Dimmable Lighting Control Module."""
from insteonplm.devices import Device
from insteonplm.states.dimmable import (
    DimmableSwitch,
    DimmableSwitch_Fan,
    DimmableKeypadA,
)
from insteonplm.states.onOff import OnOffKeypad, OnOffKeypadLed


class DimmableLightingControl(Device):
    """Dimmable Lighting Controller.

    INSTEON On/Off switch device class. Available device control options are:
        - light_on(onlevel=0xff)
        - light_on_fast(onlevel=0xff)
        - light_off()
        - light_off_fast()

    To monitor changes to the state of the device subscribe to the state
    monitor:
         - lightOnLevel.connect(callback)  (state='LightOnLevel')

    where callback defined as:
        - callback(self, device_id, state, state_value)
    """

    def __init__(
        self, plm, address, cat, subcat, product_key=None, description=None, model=None
    ):
        """Init the DimmableLightingControl Class."""
        Device.__init__(
            self, plm, address, cat, subcat, product_key, description, model
        )

        self._stateList[0x01] = DimmableSwitch(
            self._address,
            "lightOnLevel",
            0x01,
            self._send_msg,
            self._message_callbacks,
            0x00,
        )


class DimmableLightingControl_2475F(DimmableLightingControl):
    """FanLinc model 2475F Dimmable Lighting Control.

    Device Class 0x01 subcat 0x2e

    Two separate INSTEON On/Off switch devices are created with ID
        1) Ligth
            - ID: xxxxxx (where xxxxxx is the Insteon address of the device)
            - Controls:
                - light_on(onlevel=0xff)
                - light_on_fast(onlevel=0xff)
                - light_off()
                - light_off_fast()
            - Monitor: lightOnLevel.connect(callback)
        2) Fan
            - ID: xxxxxx_2  (where xxxxxx is the Insteon address of the device)
            - Controls:
                - fan_on(onlevel=0xff)
                - fan_off()
                - light_on(onlevel=0xff)  - Same as fan_on(onlevel=0xff)
                - light_off()  - Same as fan_off()
            - Monitor: fanSpeed.connect(callback)

    where callback defined as:
        - callback(self, device_id, state, state_value)
    """

    def __init__(
        self, plm, address, cat, subcat, product_key=None, description=None, model=None
    ):
        """Init the DimmableLightingControl_2475F Class."""
        super().__init__(plm, address, cat, subcat, product_key, description, model)

        self._stateList[0x01] = DimmableSwitch(
            self._address,
            "lightOnLevel",
            0x01,
            self._send_msg,
            self._message_callbacks,
            0x00,
        )
        self._stateList[0x02] = DimmableSwitch_Fan(
            self._address,
            "fanOnLevel",
            0x02,
            self._send_msg,
            self._message_callbacks,
            0x00,
        )


class DimmableLightingControl_2334_222(Device):
    """On/Off KeypadLinc Switched Lighting Control."""

    def __init__(
        self, plm, address, cat, subcat, product_key=None, description=None, model=None
    ):
        """Init the SwichedLightingControlKeypad device class."""
        super().__init__(plm, address, cat, subcat, product_key, description, model)

        self._leds = OnOffKeypadLed(
            self._address,
            "keypadLEDs",
            0x00,
            self._send_msg,
            self._message_callbacks,
            0x00,
            self._plm.loop,
        )

        self._stateList[0x01] = DimmableKeypadA(
            self._address,
            "keypadButtonMain",
            0x01,
            self._send_msg,
            self._message_callbacks,
            0x00,
            self._leds,
        )

    def _add_buttons(self, button_list):
        for group in button_list:
            self._stateList[group] = OnOffKeypad(
                self._address,
                "keypadButton{}".format(button_list[group]),
                group,
                self._send_msg,
                self._message_callbacks,
                0x00,
                self._plm.loop,
                self._leds,
            )

            self._leds.register_led_updates(self._stateList[group].led_changed, group)


class DimmableLightingControl_2334_222_8(DimmableLightingControl_2334_222):
    """Dimmable 8 Button KeypadLinc Switched Lighting Control."""

    def __init__(
        self, plm, address, cat, subcat, product_key=None, description=None, model=None
    ):
        """Init the SwitchedLightingControl_2487S device class."""
        super().__init__(plm, address, cat, subcat, product_key, description, model)

        button_list = {2: "B", 3: "C", 4: "D", 5: "E", 6: "F", 7: "G", 8: "H"}
        self._add_buttons(button_list)


class DimmableLightingControl_2334_222_6(DimmableLightingControl_2334_222):
    """Dimmable 6 Button KeypadLinc Switched Lighting Control."""

    def __init__(
        self, plm, address, cat, subcat, product_key=None, description=None, model=None
    ):
        """Init the SwitchedLightingControl_2487S device class."""
        super().__init__(plm, address, cat, subcat, product_key, description, model)

        button_list = {3: "A", 4: "B", 5: "C", 6: "D"}
        self._add_buttons(button_list)



"""INSTEON Climate Control Device Class."""
import logging

from insteonplm.devices import Device
from insteonplm.constants import COMMAND_EXTENDED_GET_SET_0X2E_0X00
from insteonplm.messages.extendedSend import ExtendedSend
from insteonplm.messages.userdata import Userdata
from insteonplm.states.thermostat import (
    Temperature,
    Humidity,
    SystemMode,
    FanMode,
    CoolSetPoint,
    HeatSetPoint,
)
from insteonplm.states.statusReport import StatusReport

_LOGGER = logging.getLogger(__name__)


class ClimateControl_2441th(Device):
    """Thermostat model 2441TH."""

    def __init__(
        self, plm, address, cat, subcat, product_key=None, description=None, model=None
    ):
        """Init the DimmableLightingControl Class."""
        Device.__init__(
            self, plm, address, cat, subcat, product_key, description, model
        )

        self._stateList[0x01] = CoolSetPoint(
            self._address,
            "coolSetPoint",
            0x01,
            self._send_msg,
            self._message_callbacks,
            0x00,
        )

        self._stateList[0x02] = HeatSetPoint(
            self._address,
            "heatSetPoint",
            0x02,
            self._send_msg,
            self._message_callbacks,
            0x00,
        )

        self._stateList[0xEF] = StatusReport(
            self._address,
            "statusReport",
            0xEF,
            self._send_msg,
            self._message_callbacks,
            0x00,
        )

        self._system_mode = SystemMode(
            self._address,
            "systemMode",
            0x10,
            self._send_msg,
            self._message_callbacks,
            0x00,
        )

        self._fan_mode = FanMode(
            self._address,
            "fanMode",
            0x11,
            self._send_msg,
            self._message_callbacks,
            0x00,
        )

        self._temp = Temperature(
            self._address,
            "temperature",
            0x12,
            self._send_msg,
            self._message_callbacks,
            0x00,
        )

        self._humidity = Humidity(
            self._address,
            "humidity",
            0x13,
            self._send_msg,
            self._message_callbacks,
            0x00,
        )

    @property
    def cool_set_point(self):
        """Return the cool set point state."""
        return self._stateList[0x01]

    @property
    def heat_set_point(self):
        """Return the heat set point state."""
        return self._stateList[0x02]

    @property
    def system_mode(self):
        """Return the mode state."""
        return self._system_mode

    @property
    def fan_mode(self):
        """Return the mode state."""
        return self._fan_mode

    @property
    def temperature(self):
        """Return the temperature state."""
        return self._temp

    @property
    def humidity(self):
        """Return the humidity state."""
        return self._humidity

    def async_refresh_state(self):
        """Request each state to provide status update."""
        _LOGGER.debug("Setting up extended status")
        ext_status = ExtendedSend(
            address=self._address,
            commandtuple=COMMAND_EXTENDED_GET_SET_0X2E_0X00,
            cmd2=0x02,
            userdata=Userdata(),
        )
        ext_status.set_crc()
        _LOGGER.debug("Sending ext status: %s", ext_status)
        self._send_msg(ext_status)
        _LOGGER.debug("Sending temp status request")
        self.temperature.async_refresh_state()

    # pylint: disable=unused-argument
    def _mode_changed(self, addr, group, val):
        self.async_refresh_state()



"""Test insteonplm.evices module."""
import asyncio
import logging

from insteonplm.constants import (COMMAND_LIGHT_OFF_0X13_0X00,
                                  COMMAND_LIGHT_ON_0X11_NONE,
                                  MESSAGE_ACK)
from insteonplm.messages.standardSend import StandardSend
from insteonplm.devices import create, DIRECT_ACK_WAIT_TIMEOUT
from insteonplm.devices.dimmableLightingControl import DimmableLightingControl
from tests.mockPLM import MockPLM

_LOGGING = logging.getLogger(__name__)
_LOGGING.setLevel(logging.DEBUG)
_INSTEON_LOGGER = logging.getLogger('insteonplm')
_INSTEON_LOGGER.setLevel(logging.DEBUG)


def test_create_device():
    """Test create device."""
    async def run_test(loop):
        plm = MockPLM(loop)
        device = create(plm, '112233', 0x01, 0x0d, None)
        assert device.id == '112233'
        assert isinstance(device, DimmableLightingControl)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_test(loop))
    open_tasks = asyncio.Task.all_tasks(loop=loop)

    for task in open_tasks:
        if hasattr(task, 'name'):
            _LOGGING.error('Device: %s Task: %s', task.name, task)
        else:
            _LOGGING.error('Task: %s', task)
        if not task.done():
            loop.run_until_complete(task)


def test_create_device_from_bytearray():
    """Test create device from byte array."""
    async def run_test(loop):
        plm = MockPLM(loop)
        target = bytearray()
        target.append(0x01)
        target.append(0x0d)
        device = create(plm, '112233', target[0], target[1], None)
        assert device.id == '112233'
        assert isinstance(device, DimmableLightingControl)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_test(loop))

    open_tasks = asyncio.Task.all_tasks(loop=loop)

    for task in open_tasks:
        if hasattr(task, 'name'):
            _LOGGING.error('Device: %s Task: %s', task.name, task)
        else:
            _LOGGING.error('Task: %s', task)
        if not task.done():
            loop.run_until_complete(task)


def test_send_msg():
    """Test sending a message."""
    async def run_test(loop):
        mockPLM = MockPLM(loop)
        address = '1a2b3c'
        device = create(mockPLM, address, 0x01, 0x0d, 0x44)
        mockPLM.devices[address] = device

        # Send the ON command. This should be sent directly to the PLM
        device.states[0x01].on()
        await asyncio.sleep(.1, loop=loop)

        # Send the OFF command. This should wait in queue until the
        # Direct ACK timeout
        device.states[0x01].off()
        await asyncio.sleep(.1, loop=loop)

        # ACK the ON command
        msgreceived = StandardSend(address, COMMAND_LIGHT_ON_0X11_NONE,
                                   cmd2=0xff, flags=0x00, acknak=MESSAGE_ACK)
        mockPLM.message_received(msgreceived)
        asyncio.sleep(.1, loop=loop)

        _LOGGING.debug('Assert that the ON command is the command in the PLM')
        assert mockPLM.sentmessage == StandardSend(
            address, COMMAND_LIGHT_ON_0X11_NONE, cmd2=0xff, flags=0x00).hex

        # Sleep until the Direct ACK time out should expire
        await asyncio.sleep(DIRECT_ACK_WAIT_TIMEOUT + .2, loop=loop)

        # Confirm that the OFF command made it to the PLM
        assert mockPLM.sentmessage == StandardSend(
            address, COMMAND_LIGHT_OFF_0X13_0X00).hex

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_test(loop))

    open_tasks = asyncio.Task.all_tasks(loop=loop)

    for task in open_tasks:
        if hasattr(task, 'name'):
            _LOGGING.error('Device: %s Task: %s', task.name, task)
        else:
            _LOGGING.error('Task: %s', task)
        if not task.done():
            loop.run_until_complete(task)



"""Insteon Powerline Modem Interface Module.

This module provides a unified asyncio network handler for interacting with
Insteon Powerline modems like the 2413U and 2412S.
"""
import asyncio
import binascii
from contextlib import suppress
import logging
import os

import aiohttp
from serial_asyncio import create_serial_connection

from insteonplm.plm import PLM, Hub

__all__ = "Connection"
_LOGGER = logging.getLogger(__name__)


async def create_http_connection(loop, protocol_factory, host, port=25105, auth=None):
    """Create an HTTP session used to connect to the Insteon Hub."""
    protocol = protocol_factory()
    transport = HttpTransport(loop, protocol, host, port, auth)
    _LOGGER.debug("create_http_connection Finished creating connection")
    return (transport, protocol)


# pylint: disable=too-many-instance-attributes
class Connection:
    """Handler to maintain the Powerline device connection."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        device=None,
        host=None,
        username=None,
        password=None,
        port=25105,
        hub_version=2,
        loop=None,
        retry_interval=1,
        auto_reconnect=True,
    ):
        """Init the Connecton class."""
        if os.name == "nt":
            self._device = device.upper()
        else:
            self._device = device
        self._host = host
        self._username = username
        self._password = password
        self._port = port
        self._hub_version = hub_version
        self._loop = loop if loop else asyncio.get_event_loop()
        self._retry_interval = retry_interval
        self._closed = True
        self._closing = False
        self._halted = False
        self._auto_reconnect = auto_reconnect
        self._protocol = None

    @property
    def protocol(self):
        """Return the connection protocol."""
        return self._protocol

    @protocol.setter
    def protocol(self, val):
        """Set the connection protocol."""
        self._protocol = val

    @property
    def device(self):
        """Return the PLM device."""
        return self._device

    @property
    def host(self):
        """Return the Hub host or IP address."""
        return self._host

    @property
    def username(self):
        """Return the Hub username."""
        return self._username

    @property
    def password(self):
        """Return the Hub password."""
        return self._password

    @property
    def port(self):
        """Return the Hub IP port."""
        return self._port

    @property
    def hub_version(self):
        """Return the version of the Insteon Hub."""
        return self._hub_version

    @property
    def auto_reconnect(self):
        """Return the connection protocol."""
        return self._auto_reconnect

    @property
    def closing(self):
        """Return the connection protocol."""
        return self._closing

    @property
    def loop(self):
        """Return the connection loop."""
        return self._loop

    # pylint: disable=too-many-arguments
    @classmethod
    async def create(
        cls,
        device="/dev/ttyUSB0",
        host=None,
        username=None,
        password=None,
        port=25010,
        hub_version=2,
        auto_reconnect=True,
        loop=None,
        workdir=None,
        poll_devices=True,
        load_aldb=True,
    ):
        """Create a connection to a specific device.

        Here is where we supply the device and callback callables we
        expect for this PLM class object.

        :param device:
            Unix device where the PLM is attached for direct connections
        :param host:
            IP Address of the Hub
        :param username:
            User name for connecting to the Hub
        :param password:
            Password for connecting to the Hub
        :param port:
            TCP port for connecting to the Hub
        :param hub_version:
            Supports version 1 (2242 serial socket) and 2 (http)
        :param auto_reconnect:
            Should the Connection try to automatically reconnect if needed?
        :param loop:
            asyncio.loop for async operation
        :param workdir:
            The directory where insteonplm should cache device discovery info
        :param poll_devices:
            Should insteonplm poll all discovered devices at startup
        :param load_aldb:
            Should the ALDB be loaded on connect

        :type device:
            str
        :type auto_reconnect:
            boolean
        :type loop:
            asyncio.loop

        """
        _LOGGER.debug("Starting Modified Connection.create")
        conn = cls(
            device=device,
            host=host,
            username=username,
            password=password,
            port=port,
            hub_version=hub_version,
            loop=loop,
            retry_interval=1,
            auto_reconnect=auto_reconnect,
        )

        def connection_lost():
            """Respond to Protocol connection lost."""
            if conn.auto_reconnect and not conn.closing:
                _LOGGER.debug("Reconnecting to transport")
                asyncio.ensure_future(conn.reconnect(), loop=conn.loop)

        protocol_class = PLM
        if conn.host and conn.hub_version == 2:
            protocol_class = Hub
        conn.protocol = protocol_class(
            connection_lost_callback=connection_lost,
            loop=conn.loop,
            workdir=workdir,
            poll_devices=poll_devices,
            load_aldb=load_aldb,
        )

        await conn.reconnect()

        _LOGGER.debug("Ending Connection.create")
        return conn

    @property
    def transport(self):
        """Return pointer to the transport object.

        Use this property to obtain passthrough access to the underlying
        Protocol properties and methods.
        """
        return self.protocol.transport

    def _reset_retry_interval(self):
        self._retry_interval = 1

    def _increase_retry_interval(self):
        self._retry_interval = min(300, 1.5 * self._retry_interval)

    async def reconnect(self):
        """Reconnect to the modem."""
        _LOGGER.debug("starting Connection.reconnect")
        await self._connect()
        while self._closed:
            await self._retry_connection()
        _LOGGER.debug("ending Connection.reconnect")

    # pylint: disable=unused-argument
    async def close(self, event):
        """Close the PLM device connection and don't try to reconnect."""
        _LOGGER.info("Closing connection to Insteon Modem")
        self._closing = True
        self._auto_reconnect = False
        await self.protocol.close()
        if self.protocol.transport:
            self.protocol.transport.close()
        await asyncio.sleep(0, loop=self._loop)
        _LOGGER.info("Insteon Modem connection closed")

    def halt(self):
        """Close the PLM device connection and wait for a resume() request."""
        _LOGGER.warning("Halting connection to Insteon Modem")
        self._halted = True
        if self.protocol.transport:
            self.protocol.transport.close()

    def resume(self):
        """Resume the PLM device connection if we have been halted."""
        _LOGGER.warning("Resuming connection to Insteon Modem")
        self._halted = False

    @property
    def dump_conndata(self):
        """Developer tool for debugging forensics."""
        attrs = vars(self)
        return ", ".join("%s: %s" % item for item in attrs.items())

    async def _retry_connection(self):
        _LOGGER.debug("starting Connection._retry_connection")
        device = self.host if self.host else self.device
        self._increase_retry_interval()
        _LOGGER.warning(
            "Connection failed, retry in %i seconds: %s", self._retry_interval, device
        )
        await asyncio.sleep(self._retry_interval, loop=self._loop)
        _LOGGER.debug("Starting _connect")
        await self._connect()
        _LOGGER.debug("ending Connection._retry_connection")

    async def _connect(self):
        _LOGGER.debug("starting Connection._connect")
        if self.host and self._hub_version == 2:
            connected = await self._connect_http()
        else:
            connected = await self._connect_serial()
        _LOGGER.debug("ending Connection._connect")
        return connected

    async def _connect_http(self):
        _LOGGER.info("Connecting to Insteon Hub on %s", self.host)
        auth = aiohttp.BasicAuth(self.username, self.password)
        _LOGGER.debug("Creating http connection")
        # pylint: disable=unused-variable
        transport, protocol = await create_http_connection(
            self._loop, lambda: self.protocol, self.host, port=self.port, auth=auth
        )
        connected = await transport.test_connection()
        if connected:
            transport.resume_reading()
        self._closed = not connected
        return connected

    async def _connect_serial(self):
        try:
            if self._hub_version == 1:
                url = "socket://{}:{}".format(self._host, self._port)
                _LOGGER.info("Connecting to Insteon Hub v1 on %s", url)
                # pylint: disable=unused-variable
                transport, protocol = await create_serial_connection(
                    self._loop, lambda: self.protocol, url, baudrate=19200
                )
            else:
                _LOGGER.info("Connecting to PLM on %s", self._device)
                # pylint: disable=unused-variable
                transport, protocol = await create_serial_connection(
                    self._loop, lambda: self.protocol, self._device, baudrate=19200
                )
            self._closed = False
        except OSError:
            self._closed = True
        return not self._closed


# Hub version 1 (2242) is untested using the HTTP Transport.
# It is tested using the PLM socket interface on port 9761.
# pylint: disable=too-many-instance-attributes
class HttpTransport(asyncio.Transport):
    """An asyncio transport model of an HTTP communication channel.

    A transport class is an abstraction of a communication channel.
    This allows protocol implementations to be developed against the
    transport abstraction without needing to know the details of the
    underlying channel, such as whether it is a pipe, a socket, or
    indeed an HTTP connection.


    You generally won’t instantiate a transport yourself; instead, you
    will call `create_http_connection` which will create the
    transport and try to initiate the underlying communication channel,
    calling you back when it succeeds.
    """

    def __init__(self, loop, protocol, host, port=25105, auth=None):
        """Init the HttpTransport class."""
        super().__init__()
        self._loop = loop
        self._protocol = protocol
        self._host = host
        self._port = port
        self._auth = auth

        self._closing = False
        self._protocol_paused = False
        self._max_read_size = 1024
        self._write_buffer = []
        self._has_reader = False
        self._has_writer = False
        self._poll_wait_time = 0.0005
        self._read_write_lock = asyncio.Lock(loop=self._loop)
        self._last_read = asyncio.Queue(loop=self._loop)
        self._restart_reader = True
        _LOGGER.debug("Starting the reader in HttpTrasnport __init__")
        self._reader_task = None

    def abort(self):
        """Abort the connection."""
        self.close()

    def can_write_eof(self):
        """Return if modem can write end of file."""
        return False

    def is_closing(self):
        """Return the state of the connection."""
        return self._closing

    def close(self):
        """Close connection to modem."""
        _LOGGER.debug("Closing Hub session")
        asyncio.ensure_future(self._close(), loop=self._loop)

    async def _close(self):
        self._closing = True
        self._restart_reader = False
        _LOGGER.info("Insteon Hub session closed")

    def get_write_buffer_size(self):
        """Get write buffer size."""
        return 0

    def pause_reading(self):
        """Pause reading."""
        asyncio.ensure_future(self._stop_reader(False), loop=self._loop)

    def resume_reading(self):
        """Resume reading."""
        self._restart_reader = True
        self._start_reader()

    def set_write_buffer_limits(self, high=None, low=None):
        """Set write buffer limits."""
        raise NotImplementedError("HTTP connections do not support write buffer limits")

    def write(self, data):
        """Write to modem."""
        _LOGGER.debug("..................Writing a message..............")
        hex_data = binascii.hexlify(data).decode()
        url = "http://{:s}:{:d}/3?{:s}=I=3".format(self._host, self._port, hex_data)
        asyncio.ensure_future(self._async_write(url), loop=self._loop)

    async def test_connection(self):
        """Test the connection to the hub."""
        url = "http://{:s}:{:d}/buffstatus.xml".format(self._host, self._port)
        response_status = 999
        try:
            async with aiohttp.ClientSession(
                loop=self._loop, auth=self._auth
            ) as session:
                async with session.get(url, timeout=10) as response:
                    if response:
                        response_status = response.status
                        if response.status == 200:
                            _LOGGER.debug(
                                "Test connection status is %d", response.status
                            )
                            return True
                        self._log_error(response.status)
                        _LOGGER.debug("Connection test failed")
                        return False

        # pylint: disable=broad-except
        except Exception as e:
            _LOGGER.error(
                "An aiohttp error occurred: %s with status %s", str(e), response_status
            )
        _LOGGER.debug("Connection test failed")
        self.close()
        return False

    async def _async_write(self, url):
        return_status = 500
        # if self._session.closed:
        #     _LOGGER.warning("Session closed, cannot write to Hub")
        #     return 999
        _LOGGER.debug("Writing message: %s", url)
        try:
            await self._read_write_lock
            async with aiohttp.ClientSession(
                loop=self._loop, auth=self._auth
            ) as session:
                async with session.post(url, timeout=10) as response:
                    return_status = response.status
                    _LOGGER.debug("Post status: %s", response.status)
                    if response.status == 200:
                        self._write_last_read(0)
                    else:
                        self._log_error(response.status)
                        await self._stop_reader(True)
        except aiohttp.client_exceptions.ServerDisconnectedError:
            _LOGGER.error("Reconnect to Hub (ServerDisconnectedError)")
            await self._stop_reader(True)
        except aiohttp.client_exceptions.ClientConnectorError:
            _LOGGER.error("Reconnect to Hub (ClientConnectorError)")
            await self._stop_reader(True)
        except asyncio.TimeoutError:
            _LOGGER.error("Reconnect to Hub (TimeoutError)")
            await self._stop_reader(True)

        if self._read_write_lock.locked():
            self._read_write_lock.release()
        return return_status

    def write_eof(self):
        """Write end of file."""
        raise NotImplementedError("HTTP connections do not support end-of-file")

    def writelines(self, list_of_data):
        """Write lines."""
        raise NotImplementedError("HTTP connections do not support writelines")

    def clear_buffer(self):
        """Clear the Hub read buffer."""
        asyncio.wait(self._clear_buffer())

    async def _clear_buffer(self):
        _LOGGER.debug("..................Clearing the buffer..............")
        url = "http://{:s}:{:d}/1?XB=M=1".format(self._host, self._port)
        await self._async_write(url)

    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements
    # pylint: disable=broad-except
    async def _ensure_reader(self):
        _LOGGER.info("Insteon Hub reader started")
        await self._clear_buffer()
        self._write_last_read(0)
        url = "http://{:s}:{:d}/buffstatus.xml".format(self._host, self._port)
        _LOGGER.debug("Calling connection made")
        _LOGGER.debug("Protocol: %s", self._protocol)
        self._protocol.connection_made(self)
        while self._restart_reader and not self._closing:
            try:
                await self._read_write_lock
                async with aiohttp.ClientSession(
                    loop=self._loop, auth=self._auth
                ) as session:
                    async with session.get(url, timeout=10) as response:
                        buffer = None
                        # _LOGGER.debug("Reader status: %d", response.status)
                        if response.status == 200:
                            html = await response.text()
                            if len(html) == 234:
                                # pylint: disable=no-value-for-parameter
                                buffer = await self._parse_buffer(html)
                            else:
                                buffer = self._parse_buffer_v1(html)
                        else:
                            self._log_error(response.status)
                            await self._stop_reader(True)
                if self._read_write_lock.locked():
                    self._read_write_lock.release()
                if buffer:
                    _LOGGER.debug("New buffer: %s", buffer)
                    bin_buffer = binascii.unhexlify(buffer)
                    self._protocol.data_received(bin_buffer)
                await asyncio.sleep(1, loop=self._loop)

            except asyncio.CancelledError:
                _LOGGER.info("Stop connection to Hub (loop stopped)")
                await self._stop_reader(False)
            except GeneratorExit:
                _LOGGER.info("Stop connection to Hub (GeneratorExit)")
                await self._stop_reader(False)
            except aiohttp.client_exceptions.ServerDisconnectedError:
                _LOGGER.error("Reconnect to Hub (ServerDisconnectedError)")
                await self._stop_reader(True)
            except aiohttp.client_exceptions.ClientConnectorError:
                _LOGGER.error("Reconnect to Hub (ClientConnectorError)")
                await self._stop_reader(True)
            except asyncio.TimeoutError:
                _LOGGER.error("Reconnect to Hub (TimeoutError)")
                await self._stop_reader(True)
            except Exception as e:
                _LOGGER.error("Restarting reading due to %s", str(e))
                await self._stop_reader(True)
        _LOGGER.info("Insteon Hub reader stopped")
        return

    async def _parse_buffer(self, html):
        last_stop = 0
        if not self._last_read.empty():
            last_stop = self._last_read.get_nowait()
        buffer = ""
        raw_text = html.replace("<response><BS>", "")
        raw_text = raw_text.replace("</BS></response>", "")
        raw_text = raw_text.strip()
        if raw_text[:199] == "0" * 200:
            # Likely the buffer was cleared
            return None
        this_stop = int(raw_text[-2:], 16)
        if this_stop > last_stop:
            _LOGGER.debug("Buffer from %d to %d", last_stop, this_stop)
            buffer = raw_text[last_stop:this_stop]
        elif this_stop < last_stop:
            _LOGGER.debug("Buffer from %d to 200 and 0 to %d", last_stop, this_stop)
            buffer_hi = raw_text[last_stop:200]
            if buffer_hi == "0" * len(buffer_hi):
                # The buffer was probably reset since the last read
                buffer_hi = ""
            buffer_low = raw_text[0:this_stop]
            buffer = "{:s}{:s}".format(buffer_hi, buffer_low)
        else:
            buffer = None
        self._write_last_read(this_stop)
        return buffer

    def _parse_buffer_v1(self, html):
        # I think this is a good idea but who knows.
        # Risk is we may be loosing the next message
        # If we do this than this method must be a coroutine
        # await self._clear_buffer()
        raw_text = html.replace("<response><BS>", "")
        raw_text = raw_text.replace("</BS></response>", "")
        raw_text = raw_text.strip()
        buffer = ""
        while raw_text:
            msg, raw_text = self._find_message(raw_text)
            if msg:
                buffer = "{}{}".format(buffer, msg)
        return buffer

    @staticmethod
    def _find_message(raw_text):
        from insteonplm.messages import iscomplete

        len_raw_text = len(raw_text)
        pos = -2
        msg_len = 4  # set to 5 because the min msg len is 4 chars or 2 bytes
        msg = None
        if raw_text == "0" * len_raw_text:
            print("Likely the buffer was cleared")
            return msg, None, False
        while pos < (len_raw_text - 2) and raw_text[pos:pos + 2] != "02":
            pos = pos + 2
        print("pos is: ", pos)
        print("len_raw_text is: ", len_raw_text)
        if pos == len_raw_text - 2:
            print("02 not found")
            return msg, None, False
        if pos > 0:
            raw_text = raw_text[pos:] + raw_text[0:pos]
        while msg_len < len_raw_text and not iscomplete(
            binascii.unhexlify(raw_text[0:msg_len])
        ):
            msg_len = msg_len + 2
        if msg_len == len_raw_text:
            print("must not be a message")
            return msg, None, False
        msg = raw_text[0:msg_len]
        raw_text = raw_text[msg_len:]
        if msg_len > len_raw_text - pos:
            # We have wrapped so there cannot be another messasge
            raw_text = None
        else:
            if pos > 0 and raw_text:
                raw_text = raw_text[len(raw_text) - pos:] + raw_text[:-pos]
        return (msg, raw_text)

    def _write_last_read(self, val):
        while not self._last_read.empty():
            self._last_read.get_nowait()
        self._last_read.put_nowait(val)

    # pylint: disable=unused-argument
    def _start_reader(self, future=None):
        if self._restart_reader:
            _LOGGER.debug("Starting the buffer reader")
            self._reader_task = asyncio.ensure_future(
                self._ensure_reader(), loop=self._loop
            )
            self._reader_task.add_done_callback(self._start_reader)

    async def _stop_reader(self, reconnect=False):
        _LOGGER.debug("Stopping the reader and reconnect is %s", reconnect)
        self._restart_reader = False
        if self._reader_task:
            self._reader_task.remove_done_callback(self._start_reader)
            self._reader_task.cancel()
            with suppress(asyncio.CancelledError):
                await self._reader_task
                await asyncio.sleep(0, loop=self._loop)
        await self._protocol.pause_writing()
        # if not self._session.closed:
        #     _LOGGER.debug('Session is open so we close it')
        #     await self._close()
        if reconnect:
            _LOGGER.debug("We want to reconnect so we do...")
            self._protocol.connection_lost(False)

    def _log_error(self, status):
        if status == 401:
            _LOGGER.error("Athentication error, check your configuration")
            _LOGGER.error("If configuration is correct and restart the Hub")
            _LOGGER.error("System must be restared to reconnect to hub")
        elif status == 404:
            _LOGGER.error(
                "Hub not found at http://%s:%d, check configuration",
                self._host,
                self._port,
            )
        elif status in range(500, 600):
            _LOGGER.error("Hub returned a server error")
            _LOGGER.error("Restart the Hub and try again")
        else:
            _LOGGER.error("An unknown error has occured")
            _LOGGER.error(
                "Check the configuration and restart the Hub and " "the application"
            )



#!/usr/bin/env python3
"""Setup for insteonplm module."""
from setuptools import setup, find_packages


def readme():
    """Return README file as a string."""
    with open('README.rst', 'r') as f:
        return f.read()


setup(
    name='insteonplm',
    version='0.16.7',
    author='David McNett',
    author_email='nugget@macnugget.org',
    url='https://github.com/nugget/python-insteonplm',
    license="MIT License",
    packages=find_packages(),
    scripts=[],
    description='Python API for controlling Insteon PowerLinc Modems',
    long_description=readme(),
    include_package_data=True,
    zip_safe=True,
    install_requires=[
        'pyserial==3.2.0',
        'pyserial-asyncio',
        'async_timeout',
        'aiohttp'
    ],
    entry_points={
        'console_scripts': ['insteonplm_monitor = insteonplm.tools:monitor',
                            'insteonplm_interactive = '
                            'insteonplm.tools:interactive']
    }
)
